searchData={"items":[{"type":"module","title":"ProcessHub","doc":"This is the main public API module for the `ProcessHub` library and it is recommended to use\nonly the functions defined in this module to interact with the `ProcessHub` library.\n\nProcessHub is a library that distributes processes within the BEAM cluster. It is designed to\nbe used as a building block for distributed applications that require process distribution\nand synchronization.","ref":"ProcessHub.html"},{"type":"function","title":"ProcessHub.await/1","doc":"This function can be used to wait for the `ProcessHub` child start or stop\nfunctions to complete.\n\nThe `await/1` function should be used with the `async_wait: true` option.\n\nKeep in mind that the `await/1` function will block the calling process until\nthe response is received. If the response is not received within the timeout\nperiod, the function will return `{:error, term()}`.","ref":"ProcessHub.html#await/1"},{"type":"function","title":"Example - ProcessHub.await/1","doc":"iex> ref = ProcessHub.start_child(:my_hub, child_spec, [async_wait: true])\n    iex> ProcessHub.await(ref)\n    {:ok, {:my_child, [{:mynode, #PID<0.123.0>}]}}","ref":"ProcessHub.html#await/1-example"},{"type":"function","title":"ProcessHub.child_lookup/2","doc":"Returns information about processes that are registered with the given `t:child_id/0`.\n\nThis function queries results from the local `ets` table and does not make any network calls.\n\nThe return results contain the `t:child_spec/0` and a list of tuples where the first element is the node\nwhere the child is started, and the second element is the `pid()` of the started child.","ref":"ProcessHub.html#child_lookup/2"},{"type":"function","title":"Example - ProcessHub.child_lookup/2","doc":"iex> {_child_spec, _node_pid_tuples} = ProcessHub.child_lookup(:my_hub, :my_child)\n    {%{id: :my_child, start: {MyProcess, :start_link, []}}, [{:mynode, #PID<0.123.0>}]}","ref":"ProcessHub.html#child_lookup/2-example"},{"type":"function","title":"ProcessHub.child_spec/1","doc":"Returns the child specification for the `ProcessHub.Initializer` supervisor.","ref":"ProcessHub.html#child_spec/1"},{"type":"function","title":"ProcessHub.get_pid/2","doc":"Returns the first pid for the given child_id.\n\nAlthough the function can be handy to quickly get the pid of the child, it is\nnot recommended to use with replication strategies as it will return the first pid only.","ref":"ProcessHub.html#get_pid/2"},{"type":"function","title":"Example - ProcessHub.get_pid/2","doc":"iex> ProcessHub.get_pid(:my_hub, :my_child)\n  #PID<0.123.0>","ref":"ProcessHub.html#get_pid/2-example"},{"type":"function","title":"ProcessHub.get_pids/2","doc":"Returns a list of pids for the given child_id.","ref":"ProcessHub.html#get_pids/2"},{"type":"function","title":"Example - ProcessHub.get_pids/2","doc":"iex> ProcessHub.get_pids(:my_hub, :my_child)\n  [#PID<0.123.0>]","ref":"ProcessHub.html#get_pids/2-example"},{"type":"function","title":"ProcessHub.is_alive?/1","doc":"Checks if the `ProcessHub` with the given `t:hub_id/0` is alive.\n\nA hub is considered alive if the `ProcessHub.Initializer` supervisor process\nis running along with the required child processes for the hub to function.","ref":"ProcessHub.html#is_alive?/1"},{"type":"function","title":"Example - ProcessHub.is_alive?/1","doc":"iex> ProcessHub.is_alive?(:not_existing)\n    false","ref":"ProcessHub.html#is_alive?/1-example"},{"type":"function","title":"ProcessHub.is_locked?/1","doc":"Checks if the `ProcessHub` with the given `t:hub_id/0` is locked.\n\nA hub is considered locked if the `ProcessHub` local event queue has a priority level\ngreater than or equal to 10. This is used to throttle the hub from processing\nany new events and conserve data integrity.","ref":"ProcessHub.html#is_locked?/1"},{"type":"function","title":"Example - ProcessHub.is_locked?/1","doc":"iex> ProcessHub.is_locked?(:my_hub)\n    false","ref":"ProcessHub.html#is_locked?/1-example"},{"type":"function","title":"ProcessHub.is_partitioned?/1","doc":"Checks if the `ProcessHub` with the given `t:hub_id/0` is in a network-partitioned state.\n\nA hub is considered partitioned if the `ProcessHub.Strategy.PartitionTolerance` strategy\nhas detected a network partition. When a network partition is detected, the hub will\nterminate the `ProcessHub.DistributedSupervisor` process along with its children.","ref":"ProcessHub.html#is_partitioned?/1"},{"type":"function","title":"Example - ProcessHub.is_partitioned?/1","doc":"iex> ProcessHub.is_partitioned?(:my_hub)\n    false","ref":"ProcessHub.html#is_partitioned?/1-example"},{"type":"function","title":"ProcessHub.nodes/2","doc":"Returns a list of nodes where the `ProcessHub` with the given `t:hub_id/0` is running.\n\nNodes where the `ProcessHub` is running with the same `t:hub_id/0` are considered\nto be part of the same cluster.","ref":"ProcessHub.html#nodes/2"},{"type":"function","title":"Example - ProcessHub.nodes/2","doc":"iex> ProcessHub.nodes(:my_hub, [:include_local])\n    [:remote_node]","ref":"ProcessHub.html#nodes/2-example"},{"type":"function","title":"ProcessHub.process_list/2","doc":"Returns a list of processes that are registered.\n\nThe process list contains the `t:child_id/0` and depending on the scope\noption, it may contain the node and `pid()` of the child.\n\nThis function queries results from the local `ets` table and does not make any network calls.\n\nAvailable options:\n- `:global` - returns a list of all child processes on all nodes in the cluster.\n  The return result will be in the format of `[{child_id, [{:node, pid}]}]`.\n- `:local` - returns a list of child processes that belong to the local node.\n  The return result will be in the format of `[{child_id, [pid]}]`\n  but only the processes that belong to the local node will be returned.","ref":"ProcessHub.html#process_list/2"},{"type":"function","title":"Example - ProcessHub.process_list/2","doc":"iex> ProcessHub.process_list(:my_hub, :global)\n    [\n      {:my_child1, [{:node1, #PID<0.123.0>}, {:node2, #PID<2.123.0>}]},\n      {:my_child2, [{:node2, #PID<5.124.0>}]}\n    ]\n    iex> ProcessHub.process_list(:my_hub, :local)\n    [{:my_child1, #PID<0.123.0>}]","ref":"ProcessHub.html#process_list/2-example"},{"type":"function","title":"ProcessHub.process_registry/1","doc":"Returns all information in the registry.\n\nThis function queries results from the local `ets` table and does not make any network calls.","ref":"ProcessHub.html#process_registry/1"},{"type":"function","title":"ProcessHub.start_child/3","doc":"Starts a child process that will be distributed across the cluster.\nThe `t:child_spec()` `:id` must be unique.","ref":"ProcessHub.html#start_child/3"},{"type":"function","title":"Example - ProcessHub.start_child/3","doc":"iex> child_spec = %{id: :my_child, start: {MyProcess, :start_link, []}}\n    iex> ProcessHub.start_child(:my_hub, child_spec)\n    {:ok, :start_initiated}\n\nBy default, the `start_child/3` function is **asynchronous** and returns immediately.\nTo wait for the child to start, you can pass `async_wait: true` to the `opts` argument.\nWhen `async_wait: true`, you **must await** the response from the function.\n\nSee `t:init_opts/0` for more options.","ref":"ProcessHub.html#start_child/3-example"},{"type":"function","title":"Example with synchronous wait - ProcessHub.start_child/3","doc":"The synchronous response includes the status code `:ok` or `:error`, a tuple containing the `t:child_id/0` and\na list of tuples where the first key is the node where the child is started, and the second key is the\n`pid()` of the started child. By default, the list should contain only one tuple, but if the\nredundancy strategy is configured for replicas, it may contain more than one tuple.\n\n    iex> child_spec = %{id: :my_child, start: {MyProcess, :start_link, []}}\n    iex> ProcessHub.start_child(:my_hub, child_spec, [async_wait: true]) |> ProcessHub.await()\n    {:ok, {:my_child, [{:mynode, #PID<0.123.0>}]}}","ref":"ProcessHub.html#start_child/3-example-with-synchronous-wait"},{"type":"function","title":"ProcessHub.start_children/3","doc":"Starts multiple child processes that will be distributed across the cluster.\n\nSame as `start_child/3`, except it starts multiple children at once and is more\nefficient than calling `start_child/3` multiple times.\n\n> #### Warning {: .warning}\n>\n> Using `start_children/3` with `async_wait: true` can lead to timeout errors,\n> especially when the number of children is large.\n> When `async_wait: true`, you **must await** the response from the function.","ref":"ProcessHub.html#start_children/3"},{"type":"function","title":"ProcessHub.start_link/1","doc":"Starts the `ProcessHub` with the given `t:hub_id/0` and settings.\n\nIt is recommended to start the `ProcessHub` under a supervision tree.","ref":"ProcessHub.html#start_link/1"},{"type":"function","title":"ProcessHub.stop/1","doc":"Stops the `ProcessHub` with the given `t:hub_id/0`.","ref":"ProcessHub.html#stop/1"},{"type":"function","title":"ProcessHub.stop_child/3","doc":"Stops a child process in the cluster.\n\nBy default, this function is **asynchronous** and returns immediately.\nYou can wait for the child to stop by passing `async_wait: true` in the `opts` argument.\nWhen `async_wait: true`, you **must await** the response from the function.","ref":"ProcessHub.html#stop_child/3"},{"type":"function","title":"Example - ProcessHub.stop_child/3","doc":"iex> ProcessHub.stop_child(:my_hub, :my_child)\n    {:ok, :stop_initiated}\n\nSee `t:stop_opts/0` for more options.","ref":"ProcessHub.html#stop_child/3-example"},{"type":"function","title":"Example with synchronous wait - ProcessHub.stop_child/3","doc":"iex> ProcessHub.stop_child(:my_hub, :my_child, [async_wait: true]) |> ProcessHub.await()\n    {:ok, {:my_child, [:mynode]}}","ref":"ProcessHub.html#stop_child/3-example-with-synchronous-wait"},{"type":"function","title":"ProcessHub.stop_children/3","doc":"Stops multiple child processes in the cluster.\n\nThis function is similar to `stop_child/3`, but it stops multiple children at once, making it more\nefficient than calling `stop_child/3` multiple times.\n\n> #### Warning {: .info}\n>\n> Using `stop_children/3` with `async_wait: true` can lead to timeout errors,\n> especially when stopping a large number of child processes.","ref":"ProcessHub.html#stop_children/3"},{"type":"function","title":"ProcessHub.which_children/2","doc":"Works similarly to `Supervisor.which_children/1`, but wraps the result in a tuple\ncontaining the node name and the children.\n\n> #### Info {: .info}\n>\n> The `Supervisor.which_children/1` function is known to consume a lot of memory\n> and this can affect performance. The problem is even more relevant when\n> using the `:global` option as it will make a network call to all nodes in the cluster.\n>\n> It is highly recommended to use `ProcessHub.process_list/2` instead.\n\nAvailable options:\n- `:global` - returns a list of all child processes started by all nodes in the cluster.\n  The return result will be in the format of `[{:node, children}]`.\n- `:local` - returns a list of all child processes started by the local node.\n  The return result will be in the format of `{:node, children}`.","ref":"ProcessHub.html#which_children/2"},{"type":"type","title":"ProcessHub.child_id/0","doc":"The `child_id` defines the name of the child. It is used to identify the child.\nEach child must have a unique `child_id()` in the cluster. A single child may have\nmultiple `pid()`s across the cluster.","ref":"ProcessHub.html#t:child_id/0"},{"type":"type","title":"ProcessHub.child_spec/0","doc":"The `child_spec` defines the specification of a child process.","ref":"ProcessHub.html#t:child_spec/0"},{"type":"type","title":"ProcessHub.hub_id/0","doc":"The `hub_id` defines the name of the hub. It is used to identify the hub.","ref":"ProcessHub.html#t:hub_id/0"},{"type":"type","title":"ProcessHub.init_opts/0","doc":"The `init_opts()` defines the options that can be passed to the `start_children/3`, `start_child/3`,\n`stop_children/3`, and `stop_child/3` functions.\n\n- `:async_wait` - is optional and is used to define whether the function should return another function that\ncan be used to wait for the children to start or stop. The default is `false`.\n- `:timeout` is optional and is used to define the timeout for the function. The timeout option\nshould be used with `async_wait: true`. The default is `5000` (5 seconds).\n- `:check_mailbox` - is optional and is used to define whether the function should clear\nthe mailbox of any existing messages that may overlap. It is recommended to keep this\noption `true` to avoid any unexpected behavior where `start_child/3` or `start_children/3`\ncall timeout but eventually the calling process receives the start responses later. These messages\nwill stay in that process's mailbox, and when the same process calls start child functions again with the\nsame `child_id()`s, it will receive the old responses.This option should be used with `async_wait: true`.\n The default is `true`.\n- `:check_existing` - is optional and is used to define whether the function should check if the children\nare already started. The default is `true`.","ref":"ProcessHub.html#t:init_opts/0"},{"type":"type","title":"ProcessHub.reply_to/0","doc":"The `reply_to` defines the `pid()`s that will receive the response from the hub\nwhen a child is started or stopped.","ref":"ProcessHub.html#t:reply_to/0"},{"type":"type","title":"ProcessHub.stop_opts/0","doc":"The `stop_opts()` defines the options that can be passed to the `stop_children/3` and `stop_child/3` functions.\n\n- `:async_wait` - is optional and is used to define whether the function should return another function that\ncan be used to wait for the children to stop. The default is `false`.\n- `:timeout` is optional and is used to define the timeout for the function. The timeout option\nshould be used with `async_wait: true`. The default is `5000` (5 seconds).","ref":"ProcessHub.html#t:stop_opts/0"},{"type":"type","title":"ProcessHub.t/0","doc":"This is the base configuration structure for the hub and has to be passed to the `start_link/1` function.\n- `:hub_id` is the name of the hub and is required.\n- `:child_specs` is optional and is used to define the child processes that will be started with the hub statically.\n- `:hooks` are optional and are used to define the hooks that can be triggered on specific events.\n- `:redundancy_strategy` is optional and is used to define the strategy for redundancy.\nThe default is `ProcessHub.Strategy.Redundancy.Singularity`.\n- `:migration_strategy` is optional and is used to define the strategy for migration.\nThe default is `ProcessHub.Strategy.Migration.ColdSwap`.\n- `:synchronization_strategy` is optional and is used to define the strategy for synchronization.\nThe default is `ProcessHub.Strategy.Synchronization.PubSub`.\n- `:partition_tolerance_strategy` is optional and is used to define the strategy for partition tolerance.\nThe default is `ProcessHub.Strategy.PartitionTolerance.Divergence`.\n- `:distribution_strategy` is optional and is used to define the strategy for process distribution.\nThe default is `ProcessHub.Strategy.Distribution.ConsistentHashing`.\n- `:hubs_discover_interval` is optional and is used to define the interval in milliseconds\nfor hubs to start the discovery process. The default is `60000` (1 minute).\n- `:deadlock_recovery_timeout` is optional and is used to define the timeout in milliseconds\nto recover from locked hub. Hub locking can happen for different reasons\nsuch as updating internal data, migrating processes or handling network partitions.\nThe default is `60000` (1 minute).","ref":"ProcessHub.html#t:t/0"},{"type":"module","title":"ProcessHub.Constant.Event","doc":"Custom events defined as macros.","ref":"ProcessHub.Constant.Event.html"},{"type":"type","title":"ProcessHub.Constant.Event.event_child_process_pid_update/0","doc":"Child process is restarted by the local supervisor.","ref":"ProcessHub.Constant.Event.html#t:event_child_process_pid_update/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_children_registration/0","doc":"Event used when a process has been registered in the ProcessHub registry.","ref":"ProcessHub.Constant.Event.html#t:event_children_registration/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_children_unregistration/0","doc":"Event used when a process has been unregistered from the ProcessHub registry.","ref":"ProcessHub.Constant.Event.html#t:event_children_unregistration/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_cluster_join/0","doc":"Event used when a node joins the ProcessHub cluster.","ref":"ProcessHub.Constant.Event.html#t:event_cluster_join/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_distribute_children/0","doc":"Event used when redistributing children to other nodes.","ref":"ProcessHub.Constant.Event.html#t:event_distribute_children/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_post_cluster_leave/0","doc":"Event used when a node leaves the ProcessHub cluster.","ref":"ProcessHub.Constant.Event.html#t:event_post_cluster_leave/0"},{"type":"type","title":"ProcessHub.Constant.Event.event_sync_remote_children/0","doc":"Event indicating that a remote node is trying to sync its processes.","ref":"ProcessHub.Constant.Event.html#t:event_sync_remote_children/0"},{"type":"type","title":"ProcessHub.Constant.Event.migration_add_event/0","doc":"Event used when external node sends migration event.","ref":"ProcessHub.Constant.Event.html#t:migration_add_event/0"},{"type":"module","title":"ProcessHub.Constant.Hook","doc":"Defines the list of hooks that can be used to extend the functionality of ProcessHub.","ref":"ProcessHub.Constant.Hook.html"},{"type":"function","title":"ProcessHub.Constant.Hook.child_process_pid_update/0","doc":"Hook triggered right after process has been restarted by local supervisor\nand the pid has been updated.","ref":"ProcessHub.Constant.Hook.html#child_process_pid_update/0"},{"type":"function","title":"ProcessHub.Constant.Hook.children_migrated/0","doc":"Hook triggered when a process is migrated to another node.","ref":"ProcessHub.Constant.Hook.html#children_migrated/0"},{"type":"function","title":"ProcessHub.Constant.Hook.coordinator_shutdown/0","doc":"Hook triggered inside the coordinator `terminate/2` function.","ref":"ProcessHub.Constant.Hook.html#coordinator_shutdown/0"},{"type":"function","title":"ProcessHub.Constant.Hook.forwarded_migration/0","doc":"Hook triggered when a process is migrated to another node.","ref":"ProcessHub.Constant.Hook.html#forwarded_migration/0"},{"type":"function","title":"ProcessHub.Constant.Hook.post_children_start/0","doc":"Hook triggered right after the children of a process are started.","ref":"ProcessHub.Constant.Hook.html#post_children_start/0"},{"type":"function","title":"ProcessHub.Constant.Hook.post_cluster_join/0","doc":"Hook triggered when a new node has joined the hub cluster and after handling\nthe node join event.","ref":"ProcessHub.Constant.Hook.html#post_cluster_join/0"},{"type":"function","title":"ProcessHub.Constant.Hook.post_cluster_leave/0","doc":"Hook triggered when a node has left the hub cluster and before handling\nthe node leave event.","ref":"ProcessHub.Constant.Hook.html#post_cluster_leave/0"},{"type":"function","title":"ProcessHub.Constant.Hook.post_nodes_redistribution/0","doc":"Hook triggered after processes are redistributed.","ref":"ProcessHub.Constant.Hook.html#post_nodes_redistribution/0"},{"type":"function","title":"ProcessHub.Constant.Hook.pre_children_redistribution/0","doc":"Hook triggered before redistribution of children is called.\n\nThis is only called with node addition or removal from the cluster.","ref":"ProcessHub.Constant.Hook.html#pre_children_redistribution/0"},{"type":"function","title":"ProcessHub.Constant.Hook.pre_children_start/0","doc":"Hook triggered before the children of a process are started.","ref":"ProcessHub.Constant.Hook.html#pre_children_start/0"},{"type":"function","title":"ProcessHub.Constant.Hook.pre_cluster_join/0","doc":"Hook triggered when a new node has joined the hub cluster and before handling\nthe node join event.","ref":"ProcessHub.Constant.Hook.html#pre_cluster_join/0"},{"type":"function","title":"ProcessHub.Constant.Hook.pre_cluster_leave/0","doc":"Hook triggered when a node has left the hub cluster and before handling\nthe node leave event.","ref":"ProcessHub.Constant.Hook.html#pre_cluster_leave/0"},{"type":"function","title":"ProcessHub.Constant.Hook.pre_nodes_redistribution/0","doc":"Hook triggered before processes are redistributed.","ref":"ProcessHub.Constant.Hook.html#pre_nodes_redistribution/0"},{"type":"function","title":"ProcessHub.Constant.Hook.priority_state_updated/0","doc":"Hook triggered when the priority level of the local event queue has been updated.","ref":"ProcessHub.Constant.Hook.html#priority_state_updated/0"},{"type":"function","title":"ProcessHub.Constant.Hook.process_startups/0","doc":"Hook triggered right after processes are started.","ref":"ProcessHub.Constant.Hook.html#process_startups/0"},{"type":"function","title":"ProcessHub.Constant.Hook.registry_pid_inserted/0","doc":"Hook triggered when a new process is registered in the ProcessHub registry.","ref":"ProcessHub.Constant.Hook.html#registry_pid_inserted/0"},{"type":"function","title":"ProcessHub.Constant.Hook.registry_pid_removed/0","doc":"Hook triggered when a process is unregistered from the ProcessHub registry.","ref":"ProcessHub.Constant.Hook.html#registry_pid_removed/0"},{"type":"module","title":"ProcessHub.Constant.PriorityLevel","doc":"Priority levels for the local event queue.","ref":"ProcessHub.Constant.PriorityLevel.html"},{"type":"function","title":"ProcessHub.Constant.PriorityLevel.locked/0","doc":"Priority level for the local event queue when it is locked.","ref":"ProcessHub.Constant.PriorityLevel.html#locked/0"},{"type":"function","title":"ProcessHub.Constant.PriorityLevel.unlocked/0","doc":"Priority level for the local event queue when it is unlocked.","ref":"ProcessHub.Constant.PriorityLevel.html#unlocked/0"},{"type":"type","title":"ProcessHub.Constant.PriorityLevel.priority_level/0","doc":"","ref":"ProcessHub.Constant.PriorityLevel.html#t:priority_level/0"},{"type":"module","title":"ProcessHub.Constant.StorageKey","doc":"","ref":"ProcessHub.Constant.StorageKey.html"},{"type":"function","title":"ProcessHub.Constant.StorageKey.dlrt/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#dlrt/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.dqdn/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#dqdn/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.gct/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#gct/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.gdc/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#gdc/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.hdi/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#hdi/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.hn/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#hn/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.hodw/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#hodw/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.hr/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#hr/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.msk/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#msk/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.staticcs/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#staticcs/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.strdist/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#strdist/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.strmigr/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#strmigr/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.strpart/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#strpart/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.strred/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#strred/0"},{"type":"function","title":"ProcessHub.Constant.StorageKey.strsyn/0","doc":"","ref":"ProcessHub.Constant.StorageKey.html#strsyn/0"},{"type":"module","title":"ProcessHub.Coordinator","doc":"The `ProcessHub` coordinator module is responsible for coordinating most of the `ProcessHub` events and work.\n\nIn most cases, the coordinator module delegates the work to other service-based\nmodules or handler processes that are created on demand.\n\nEach `ProcessHub` instance has its own coordinator process that handles the\ncoordination. These processes are supervised by the `ProcessHub.Initializer`\nsupervisor.\n\nThe coordinator stores state about the `ProcessHub` instance, such as the cluster nodes.\n\nAdditionally, the coordinator takes care of any periodic tasks required by the\n`ProcessHub` instance, such as initial synchronization, propagation, etc.","ref":"ProcessHub.Coordinator.html"},{"type":"function","title":"ProcessHub.Coordinator.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.Coordinator.html#child_spec/1"},{"type":"function","title":"ProcessHub.Coordinator.start_link/1","doc":"","ref":"ProcessHub.Coordinator.html#start_link/1"},{"type":"type","title":"ProcessHub.Coordinator.t/0","doc":"","ref":"ProcessHub.Coordinator.html#t:t/0"},{"type":"behaviour","title":"ProcessHub.DistributedSupervisor","doc":"The `ProcessHub` distributed supervisor module is responsible for starting and stopping\nthe child processes distributed across the cluster.\n\nEach `ProcessHub` instance has its own distributed supervisor that manages local\nchild processes.","ref":"ProcessHub.DistributedSupervisor.html"},{"type":"function","title":"ProcessHub.DistributedSupervisor.behaviour_info/1","doc":"","ref":"ProcessHub.DistributedSupervisor.html#behaviour_info/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.check_childspecs/1","doc":"","ref":"ProcessHub.DistributedSupervisor.html#check_childspecs/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.check_childspecs/2","doc":"","ref":"ProcessHub.DistributedSupervisor.html#check_childspecs/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.code_change/3","doc":"","ref":"ProcessHub.DistributedSupervisor.html#code_change/3"},{"type":"function","title":"ProcessHub.DistributedSupervisor.count_children/1","doc":"","ref":"ProcessHub.DistributedSupervisor.html#count_children/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.delete_child/2","doc":"","ref":"ProcessHub.DistributedSupervisor.html#delete_child/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.format_log/1","doc":"","ref":"ProcessHub.DistributedSupervisor.html#format_log/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.format_log/2","doc":"","ref":"ProcessHub.DistributedSupervisor.html#format_log/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.get_callback_module/1","doc":"","ref":"ProcessHub.DistributedSupervisor.html#get_callback_module/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.get_childspec/2","doc":"","ref":"ProcessHub.DistributedSupervisor.html#get_childspec/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.handle_call/3","doc":"","ref":"ProcessHub.DistributedSupervisor.html#handle_call/3"},{"type":"function","title":"ProcessHub.DistributedSupervisor.handle_cast/2","doc":"","ref":"ProcessHub.DistributedSupervisor.html#handle_cast/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.handle_info/2","doc":"Handles the process exit messages for the child processes.\n\nWe delegate the work the the `:supervisor.handle_info/2` function and then\npropagate the event to the `Dispatcher` module to notify the other nodes\nabout the child process failure.","ref":"ProcessHub.DistributedSupervisor.html#handle_info/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.has_child?/2","doc":"Returns `true` if the child process is running on local node.","ref":"ProcessHub.DistributedSupervisor.html#has_child?/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.init/1","doc":"Initializes the distributed supervisor with the given arguments.\n\nWe will call the Erlang `:supervisor.init/1` function with the given arguments.\nthis in turn will call the `init/1`","ref":"ProcessHub.DistributedSupervisor.html#init/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.local_child_ids/1","doc":"Returns the child process ids that are running on local node.","ref":"ProcessHub.DistributedSupervisor.html#local_child_ids/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.local_children/1","doc":"Returns a list of processe pairs in the form of `{child_id, pid}`\n  that are running on local node.","ref":"ProcessHub.DistributedSupervisor.html#local_children/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.local_pid/2","doc":"Returns the child process pid if it is running on local node.","ref":"ProcessHub.DistributedSupervisor.html#local_pid/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.restart_child/2","doc":"","ref":"ProcessHub.DistributedSupervisor.html#restart_child/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.start_child/2","doc":"Starts a child process on local node.","ref":"ProcessHub.DistributedSupervisor.html#start_child/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.start_link/1","doc":"Starts the distributed supervisor with the given arguments.\n\nWe will call the `GenServer.start_link/3` and register our current module\nas the base module, this way we can overwrite the `handle_info/2` function\non the `:supervisor` module to handle the `:EXIT` messages our selves.","ref":"ProcessHub.DistributedSupervisor.html#start_link/1"},{"type":"function","title":"ProcessHub.DistributedSupervisor.terminate/2","doc":"","ref":"ProcessHub.DistributedSupervisor.html#terminate/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.terminate_child/2","doc":"Stops a child process on local node by first terminating the process and then\ndeleting it from the supervisor child spec list.","ref":"ProcessHub.DistributedSupervisor.html#terminate_child/2"},{"type":"function","title":"ProcessHub.DistributedSupervisor.which_children/1","doc":"","ref":"ProcessHub.DistributedSupervisor.html#which_children/1"},{"type":"module","title":"ProcessHub.Handler.ChildrenAdd.PostStartData","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.PostStartData.html"},{"type":"type","title":"ProcessHub.Handler.ChildrenAdd.PostStartData.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.PostStartData.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenAdd.StartHandle","doc":"Handler for starting child processes.","ref":"ProcessHub.Handler.ChildrenAdd.StartHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenAdd.StartHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.StartHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenAdd.StartHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.StartHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenAdd.SyncHandle","doc":"Handler for synchronizing added child processes.","ref":"ProcessHub.Handler.ChildrenAdd.SyncHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenAdd.SyncHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.SyncHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenAdd.SyncHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenAdd.SyncHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenRem.StopHandle","doc":"Handler for stopping child processes.","ref":"ProcessHub.Handler.ChildrenRem.StopHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenRem.StopHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenRem.StopHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenRem.StopHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenRem.StopHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ChildrenRem.SyncHandle","doc":"Handler for synchronizing stopped child processes.","ref":"ProcessHub.Handler.ChildrenRem.SyncHandle.html"},{"type":"function","title":"ProcessHub.Handler.ChildrenRem.SyncHandle.handle/1","doc":"","ref":"ProcessHub.Handler.ChildrenRem.SyncHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ChildrenRem.SyncHandle.t/0","doc":"","ref":"ProcessHub.Handler.ChildrenRem.SyncHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ClusterUpdate.NodeDown","doc":"Handler for the node down event.","ref":"ProcessHub.Handler.ClusterUpdate.NodeDown.html"},{"type":"function","title":"ProcessHub.Handler.ClusterUpdate.NodeDown.handle/1","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeDown.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ClusterUpdate.NodeDown.t/0","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeDown.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.ClusterUpdate.NodeUp","doc":"Handler for the node up event.","ref":"ProcessHub.Handler.ClusterUpdate.NodeUp.html"},{"type":"function","title":"ProcessHub.Handler.ClusterUpdate.NodeUp.handle/1","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeUp.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.ClusterUpdate.NodeUp.t/0","doc":"","ref":"ProcessHub.Handler.ClusterUpdate.NodeUp.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.Synchronization.IntervalSyncHandle","doc":"Handler for periodic synchronization.","ref":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.html"},{"type":"function","title":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.handle/1","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.t/0","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.Synchronization.IntervalSyncInit","doc":"Handler for initializing synchronization.","ref":"ProcessHub.Handler.Synchronization.IntervalSyncInit.html"},{"type":"function","title":"ProcessHub.Handler.Synchronization.IntervalSyncInit.handle/1","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncInit.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.Synchronization.IntervalSyncInit.t/0","doc":"","ref":"ProcessHub.Handler.Synchronization.IntervalSyncInit.html#t:t/0"},{"type":"module","title":"ProcessHub.Handler.Synchronization.ProcessEmitHandle","doc":"Handler for emitting process registry data.","ref":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.html"},{"type":"function","title":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.handle/1","doc":"","ref":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.html#handle/1"},{"type":"type","title":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.t/0","doc":"","ref":"ProcessHub.Handler.Synchronization.ProcessEmitHandle.html#t:t/0"},{"type":"module","title":"ProcessHub.Initializer","doc":"The main `ProcessHub` initializer supervisor module.\n\nThis module is responsible for starting all the child processes of `ProcessHub`.","ref":"ProcessHub.Initializer.html"},{"type":"function","title":"ProcessHub.Initializer.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.Initializer.html#child_spec/1"},{"type":"function","title":"ProcessHub.Initializer.start_link/1","doc":"Starts a `ProcessHub` instance with all its children.","ref":"ProcessHub.Initializer.html#start_link/1"},{"type":"function","title":"ProcessHub.Initializer.stop/1","doc":"Starts a `ProcessHub` instance with all its children.","ref":"ProcessHub.Initializer.html#stop/1"},{"type":"module","title":"ProcessHub.Injector","doc":"","ref":"ProcessHub.Injector.html"},{"type":"module","title":"ProcessHub.Janitor","doc":"","ref":"ProcessHub.Janitor.html"},{"type":"function","title":"ProcessHub.Janitor.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.Janitor.html#child_spec/1"},{"type":"function","title":"ProcessHub.Janitor.start_link/1","doc":"","ref":"ProcessHub.Janitor.html#start_link/1"},{"type":"module","title":"ProcessHub.Service.Cluster","doc":"`ProcessHub` instances with the same `hub_id` will automatically form a cluster.\nThe cluster service provides API functions for managing the cluster.","ref":"ProcessHub.Service.Cluster.html"},{"type":"function","title":"ProcessHub.Service.Cluster.add_hub_node/2","doc":"Adds a new node to the hub cluster and returns new list of nodes.","ref":"ProcessHub.Service.Cluster.html#add_hub_node/2"},{"type":"function","title":"ProcessHub.Service.Cluster.new_node?/2","doc":"Returns a boolean indicating whether the node exists in the cluster.","ref":"ProcessHub.Service.Cluster.html#new_node?/2"},{"type":"function","title":"ProcessHub.Service.Cluster.nodes/2","doc":"Returns a list of nodes in the cluster.","ref":"ProcessHub.Service.Cluster.html#nodes/2"},{"type":"function","title":"ProcessHub.Service.Cluster.propagate_self/2","doc":"Sends a cluster join event to the remote node.","ref":"ProcessHub.Service.Cluster.html#propagate_self/2"},{"type":"function","title":"ProcessHub.Service.Cluster.rem_hub_node/2","doc":"Removes a node from the cluster and returns new list of nodes.","ref":"ProcessHub.Service.Cluster.html#rem_hub_node/2"},{"type":"module","title":"ProcessHub.Service.Dispatcher","doc":"The dispatcher service provides API functions for dispatching events.","ref":"ProcessHub.Service.Dispatcher.html"},{"type":"function","title":"ProcessHub.Service.Dispatcher.children_migrate/3","doc":"Sends the coordinator process a message to start the child processes passed in.","ref":"ProcessHub.Service.Dispatcher.html#children_migrate/3"},{"type":"function","title":"ProcessHub.Service.Dispatcher.children_start/3","doc":"Sends the coordinator process a message to start the child processes passed in.\n\n@TODO: should use the underlying PubSub or Gossip.","ref":"ProcessHub.Service.Dispatcher.html#children_start/3"},{"type":"function","title":"ProcessHub.Service.Dispatcher.children_stop/2","doc":"Sends the coordinator process a message to stop the child processes passed in.","ref":"ProcessHub.Service.Dispatcher.html#children_stop/2"},{"type":"function","title":"ProcessHub.Service.Dispatcher.propagate_event/4","doc":"Propagates the event to the event queue.","ref":"ProcessHub.Service.Dispatcher.html#propagate_event/4"},{"type":"function","title":"ProcessHub.Service.Dispatcher.reply_respondents/5","doc":"Sends a message to all the respondents who are waiting for a response.","ref":"ProcessHub.Service.Dispatcher.html#reply_respondents/5"},{"type":"module","title":"ProcessHub.Service.Distributor","doc":"The distributor service provides API functions for distributing child processes.","ref":"ProcessHub.Service.Distributor.html"},{"type":"function","title":"ProcessHub.Service.Distributor.children_redist_init/4","doc":"Initiates process redistribution.","ref":"ProcessHub.Service.Distributor.html#children_redist_init/4"},{"type":"function","title":"ProcessHub.Service.Distributor.children_terminate/4","doc":"Terminates child processes locally and propagates all nodes in the cluster\nto remove the child processes from their registry.","ref":"ProcessHub.Service.Distributor.html#children_terminate/4"},{"type":"function","title":"ProcessHub.Service.Distributor.init_children/3","doc":"Initiates processes startup.","ref":"ProcessHub.Service.Distributor.html#init_children/3"},{"type":"function","title":"ProcessHub.Service.Distributor.stop_children/3","doc":"Initiates processes shutdown.","ref":"ProcessHub.Service.Distributor.html#stop_children/3"},{"type":"function","title":"ProcessHub.Service.Distributor.which_children_global/2","doc":"Return a list of all child processes started by all nodes in the cluster.","ref":"ProcessHub.Service.Distributor.html#which_children_global/2"},{"type":"function","title":"ProcessHub.Service.Distributor.which_children_local/2","doc":"Returns all child processes started by the local node.\n\nWorks similar to `Supervisor.which_children/1` but returns a list of tuple\nwhere the first element is the node name and the second child processes started under the node.","ref":"ProcessHub.Service.Distributor.html#which_children_local/2"},{"type":"module","title":"ProcessHub.Service.HookManager","doc":"The hook manager service provides API functions for managing hook dispatching,\nregistration, and lookup.","ref":"ProcessHub.Service.HookManager.html"},{"type":"function","title":"ProcessHub.Service.HookManager.cancel_handler/3","doc":"Cancels a hook handler.","ref":"ProcessHub.Service.HookManager.html#cancel_handler/3"},{"type":"function","title":"ProcessHub.Service.HookManager.dispatch_hook/3","doc":"Dispatches the hook to the registered handlers and passes the hook data as an argument.\n\nIt is possible to register a hook handler with a wildcard argument `:_` which\nwill be replaced with the hook data when the hook is dispatched.","ref":"ProcessHub.Service.HookManager.html#dispatch_hook/3"},{"type":"function","title":"ProcessHub.Service.HookManager.dispatch_hooks/2","doc":"Dispatches multiple hooks to the registered handlers.","ref":"ProcessHub.Service.HookManager.html#dispatch_hooks/2"},{"type":"function","title":"ProcessHub.Service.HookManager.register_handler/3","doc":"Registers a new hook handler.","ref":"ProcessHub.Service.HookManager.html#register_handler/3"},{"type":"function","title":"ProcessHub.Service.HookManager.register_handlers/3","doc":"Registers a new hook handlers.","ref":"ProcessHub.Service.HookManager.html#register_handlers/3"},{"type":"function","title":"ProcessHub.Service.HookManager.registered_handlers/2","doc":"Returns all registered hook handlers for the given hook key","ref":"ProcessHub.Service.HookManager.html#registered_handlers/2"},{"type":"type","title":"ProcessHub.Service.HookManager.handler_id/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:handler_id/0"},{"type":"type","title":"ProcessHub.Service.HookManager.handler_priority/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:handler_priority/0"},{"type":"type","title":"ProcessHub.Service.HookManager.hook_handlers/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:hook_handlers/0"},{"type":"type","title":"ProcessHub.Service.HookManager.hook_key/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:hook_key/0"},{"type":"type","title":"ProcessHub.Service.HookManager.t/0","doc":"","ref":"ProcessHub.Service.HookManager.html#t:t/0"},{"type":"module","title":"ProcessHub.Service.Mailbox","doc":"The messenger service provides API functions for receiving messages from other processes.","ref":"ProcessHub.Service.Mailbox.html"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_child_resp/5","doc":"Waits for multiple child response messages.","ref":"ProcessHub.Service.Mailbox.html#receive_child_resp/5"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_response/3","doc":"Receives a single child response message.","ref":"ProcessHub.Service.Mailbox.html#receive_response/3"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_response/6","doc":"Receives a single child response message.","ref":"ProcessHub.Service.Mailbox.html#receive_response/6"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_start_resp/2","doc":"Waits for multiple child process startup results.","ref":"ProcessHub.Service.Mailbox.html#receive_start_resp/2"},{"type":"function","title":"ProcessHub.Service.Mailbox.receive_stop_resp/2","doc":"Waits for multiple child process termination results.","ref":"ProcessHub.Service.Mailbox.html#receive_stop_resp/2"},{"type":"module","title":"ProcessHub.Service.ProcessRegistry","doc":"The process registry service provides API functions for managing the process registry.","ref":"ProcessHub.Service.ProcessRegistry.html"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.bulk_delete/2","doc":"Deletes information about multiple child processes from the registry.\n\nCalling this function will dispatch the `:registry_pid_remove_hook` hooks unless\nthe `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#bulk_delete/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.bulk_insert/2","doc":"Inserts information about multiple child processes into the registry.\n\nCalling this function will dispatch the `:registry_pid_insert_hook` hook unless the `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#bulk_insert/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.clear_all/1","doc":"Deletes all objects from the process registry.","ref":"ProcessHub.Service.ProcessRegistry.html#clear_all/1"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.contains_children/2","doc":"Returns a list of child_ids that match the given `child_ids` variable.","ref":"ProcessHub.Service.ProcessRegistry.html#contains_children/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.delete/3","doc":"Deletes information about a child process from the registry.\n\nCalling this function will dispatch the `:registry_pid_remove_hook` hook unless the `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#delete/3"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.get_pid/2","doc":"Returns the first pid for the given child_id.","ref":"ProcessHub.Service.ProcessRegistry.html#get_pid/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.get_pids/2","doc":"Returns a list of pids for the given child_id.","ref":"ProcessHub.Service.ProcessRegistry.html#get_pids/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.insert/4","doc":"Inserts information about a child process into the registry.\n\nCalling this function will dispatch the `:registry_pid_insert_hook` hook unless the `:skip_hooks` option is set to `true`.","ref":"ProcessHub.Service.ProcessRegistry.html#insert/4"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.local_child_specs/1","doc":"Returns a list of child specs registered under the local node.","ref":"ProcessHub.Service.ProcessRegistry.html#local_child_specs/1"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.local_data/1","doc":"Returns information on all processes that are running on the local node.","ref":"ProcessHub.Service.ProcessRegistry.html#local_data/1"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.lookup/2","doc":"","ref":"ProcessHub.Service.ProcessRegistry.html#lookup/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.lookup/3","doc":"Return the child_spec, nodes, and pids for the given child_id.","ref":"ProcessHub.Service.ProcessRegistry.html#lookup/3"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.process_list/2","doc":"","ref":"ProcessHub.Service.ProcessRegistry.html#process_list/2"},{"type":"function","title":"ProcessHub.Service.ProcessRegistry.registry/1","doc":"Returns information about all registered processes.","ref":"ProcessHub.Service.ProcessRegistry.html#registry/1"},{"type":"type","title":"ProcessHub.Service.ProcessRegistry.registry/0","doc":"","ref":"ProcessHub.Service.ProcessRegistry.html#t:registry/0"},{"type":"module","title":"ProcessHub.Service.Ring","doc":"The Ring service provides API functions for managing the hash ring.","ref":"ProcessHub.Service.Ring.html"},{"type":"function","title":"ProcessHub.Service.Ring.add_node/2","doc":"Adds a new node to the passed-in `hash_ring` and returns the new hash ring.","ref":"ProcessHub.Service.Ring.html#add_node/2"},{"type":"function","title":"ProcessHub.Service.Ring.create_ring/1","doc":"Creates a new hash ring instance from the given `hub_nodes`\nand returns the new hash ring.","ref":"ProcessHub.Service.Ring.html#create_ring/1"},{"type":"function","title":"ProcessHub.Service.Ring.get_ring/1","doc":"Returns the hash ring instance belonging to the given `hub_id`.","ref":"ProcessHub.Service.Ring.html#get_ring/1"},{"type":"function","title":"ProcessHub.Service.Ring.key_to_node/3","doc":"Determines which node the given `child_id` belongs to.","ref":"ProcessHub.Service.Ring.html#key_to_node/3"},{"type":"function","title":"ProcessHub.Service.Ring.key_to_nodes/3","doc":"Determines which nodes the given `child_id` belongs to.\n\nThe `replication_factor` determines how many nodes to return.","ref":"ProcessHub.Service.Ring.html#key_to_nodes/3"},{"type":"function","title":"ProcessHub.Service.Ring.nodes/1","doc":"Returns a list of all nodes in the hash ring.","ref":"ProcessHub.Service.Ring.html#nodes/1"},{"type":"function","title":"ProcessHub.Service.Ring.remove_node/2","doc":"Removes a node from the hash ring and returns the new hash ring.","ref":"ProcessHub.Service.Ring.html#remove_node/2"},{"type":"module","title":"ProcessHub.Service.State","doc":"The state service provides API functions for managing the state of the hub and\nlocking/unlocking the local event handler.","ref":"ProcessHub.Service.State.html"},{"type":"function","title":"ProcessHub.Service.State.is_locked?/1","doc":"Returns a boolean indicating whether the hub is locked.","ref":"ProcessHub.Service.State.html#is_locked?/1"},{"type":"function","title":"ProcessHub.Service.State.is_partitioned?/1","doc":"Returns a boolean indicating whether the hub cluster is partitioned.","ref":"ProcessHub.Service.State.html#is_partitioned?/1"},{"type":"function","title":"ProcessHub.Service.State.lock_event_handler/2","doc":"Locks the event handler and dispatches the priority state updated hook.","ref":"ProcessHub.Service.State.html#lock_event_handler/2"},{"type":"function","title":"ProcessHub.Service.State.toggle_quorum_failure/1","doc":"Locks the event handler and kills the local distributed supervisor.","ref":"ProcessHub.Service.State.html#toggle_quorum_failure/1"},{"type":"function","title":"ProcessHub.Service.State.toggle_quorum_success/1","doc":"Unlocks the local event handler and restarts the local distributed supervisor.","ref":"ProcessHub.Service.State.html#toggle_quorum_success/1"},{"type":"function","title":"ProcessHub.Service.State.unlock_event_handler/1","doc":"Unlocks the event handler and dispatches the priority state updated hook.","ref":"ProcessHub.Service.State.html#unlock_event_handler/1"},{"type":"module","title":"ProcessHub.Service.Storage","doc":"The local storage service provides API functions for managing local storage.\nThis is mainly used for storing data that is not required to be persisted.\n\nLocal storage is implemented using ETS tables and is cleared periodically.","ref":"ProcessHub.Service.Storage.html"},{"type":"function","title":"ProcessHub.Service.Storage.clear_all/1","doc":"Deletes all objects from the ETS table.\n\nNever use in custom code. Should only be used for testing purposes.","ref":"ProcessHub.Service.Storage.html#clear_all/1"},{"type":"function","title":"ProcessHub.Service.Storage.exists?/2","doc":"Returns a boolean indicating whether the key exists in local storage.","ref":"ProcessHub.Service.Storage.html#exists?/2"},{"type":"function","title":"ProcessHub.Service.Storage.export_all/1","doc":"Exports all objects from the ETS table.","ref":"ProcessHub.Service.Storage.html#export_all/1"},{"type":"function","title":"ProcessHub.Service.Storage.get/2","doc":"Returns the value for the key in local storage.\n\nIf the key does not exist, nil is returned; otherwise, the value is returned in\nthe format of a tuple: `{key, value, ttl}`.","ref":"ProcessHub.Service.Storage.html#get/2"},{"type":"function","title":"ProcessHub.Service.Storage.insert/4","doc":"Inserts the key and value into local storage.\n\nAvailable options:\n  - `ttl`: The time to live for the key in milliseconds.","ref":"ProcessHub.Service.Storage.html#insert/4"},{"type":"function","title":"ProcessHub.Service.Storage.remove/2","doc":"Removes an entry from the storage.","ref":"ProcessHub.Service.Storage.html#remove/2"},{"type":"function","title":"ProcessHub.Service.Storage.update/3","doc":"Updates the value for the key in local storage.\n\nThe `func` function is expected to take the current value as an argument and\nreturn the new value.\n\nIf the key does not exist, the current value is nil.","ref":"ProcessHub.Service.Storage.html#update/3"},{"type":"type","title":"ProcessHub.Service.Storage.table_id/0","doc":"","ref":"ProcessHub.Service.Storage.html#t:table_id/0"},{"type":"module","title":"ProcessHub.Service.Synchronizer","doc":"The synchronizer service provides API functions for synchronizing process\nregistry data between nodes.","ref":"ProcessHub.Service.Synchronizer.html"},{"type":"function","title":"ProcessHub.Service.Synchronizer.append_data/2","doc":"Appends remote data to the local process registry.","ref":"ProcessHub.Service.Synchronizer.html#append_data/2"},{"type":"function","title":"ProcessHub.Service.Synchronizer.detach_data/2","doc":"Detaches remote data from the local process registry.","ref":"ProcessHub.Service.Synchronizer.html#detach_data/2"},{"type":"function","title":"ProcessHub.Service.Synchronizer.exec_interval_sync/4","doc":"","ref":"ProcessHub.Service.Synchronizer.html#exec_interval_sync/4"},{"type":"function","title":"ProcessHub.Service.Synchronizer.local_sync_data/1","doc":"Returns local node's process registry data used for synchronization.","ref":"ProcessHub.Service.Synchronizer.html#local_sync_data/1"},{"type":"protocol","title":"ProcessHub.Strategy.Distribution.Base","doc":"The distribution strategy protocol defines behaviour to identify the nodes\nwhich are responsible for a child process.","ref":"ProcessHub.Strategy.Distribution.Base.html"},{"type":"function","title":"ProcessHub.Strategy.Distribution.Base.belongs_to/4","doc":"Returns the list of nodes where the child process belongs to.\n\nThe list of nodes is used to determine where the child process should be started\nor migrated to.","ref":"ProcessHub.Strategy.Distribution.Base.html#belongs_to/4"},{"type":"function","title":"ProcessHub.Strategy.Distribution.Base.children_init/4","doc":"Perform any necessary initialization and validation for the started children.","ref":"ProcessHub.Strategy.Distribution.Base.html#children_init/4"},{"type":"function","title":"ProcessHub.Strategy.Distribution.Base.init/2","doc":"Triggered when coordinator is initialized.\n\nCould be used to perform initialization.","ref":"ProcessHub.Strategy.Distribution.Base.html#init/2"},{"type":"type","title":"ProcessHub.Strategy.Distribution.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.Distribution.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Distribution.ConsistentHashing","doc":"Provides implementation for distribution behaviour using consistent hashing.\n\nThis strategy uses consistent hashing to determine the nodes and child processes\nmapping.\n\nThe consensus is achieved in the cluster by creating a hash ring. The hash ring\nis a ring of nodes where each node is responsible for a range of hash values.\nThe hash value of a child process is used to determine which node is responsible\nfor the child process.\n\nWhen the cluster is updated, the hash ring is recalculated.\nThe recalculation is done in a way that each node is assigned a unique\nhash value, and they form a **hash ring**. Each node in the cluster keeps track of\nthe ProcessHub cluster and updates its local hash ring accordingly.\n\nTo find the node that the process belongs to, the system will use the hash ring to calculate\nthe hash value of the process ID (`child_id`) and assign it to the node with the closest hash value.\n\nWhen the cluster is updated and the hash ring is recalculated, it does not mean that all\nprocesses will be shuffled. Only the processes that are affected by the change will\nbe redistributed. This is done to avoid unnecessary process migrations.\n\nFor example, when a node leaves the cluster, only the processes that were running on that node\nwill be redistributed. The rest of the processes will stay on the same node. When a new node\njoins the cluster, only some of the processes will be redistributed to the new node, and the\nrest will stay on the same node.\n\n> The hash ring implementation **does not guarantee** that all processes will always be\n> evenly distributed, but it does its best to distribute them as evenly as possible.\n\nThis is the default distribution strategy.","ref":"ProcessHub.Strategy.Distribution.ConsistentHashing.html"},{"type":"function","title":"ProcessHub.Strategy.Distribution.ConsistentHashing.handle_node_join/2","doc":"Adds a new node to the hash ring.","ref":"ProcessHub.Strategy.Distribution.ConsistentHashing.html#handle_node_join/2"},{"type":"function","title":"ProcessHub.Strategy.Distribution.ConsistentHashing.handle_node_leave/2","doc":"Removes the node from the hash ring.","ref":"ProcessHub.Strategy.Distribution.ConsistentHashing.html#handle_node_leave/2"},{"type":"function","title":"ProcessHub.Strategy.Distribution.ConsistentHashing.handle_shutdown/1","doc":"Removes the local node from the hash ring.","ref":"ProcessHub.Strategy.Distribution.ConsistentHashing.html#handle_shutdown/1"},{"type":"type","title":"ProcessHub.Strategy.Distribution.ConsistentHashing.t/0","doc":"","ref":"ProcessHub.Strategy.Distribution.ConsistentHashing.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Distribution.Guided","doc":"Provides implementation for distribution behaviour using consistent hashing.\n\nThis strategy expects the child mappings to be provided during process\nstart-up initialization. The child mappings are used to determine the nodes and\nprocesses mapping.\n\nWhen processes are started using Guided distribution strategy, they will be\nbound to the nodes specified in the child mappings. If the node goes down, the\nprocesses won't be migrated to other nodes. They will stay on the same node\nuntil the node is restarted.\n\nGuided distribution strategy is useful when you want to have more control over\nthe distribution of processes. For example, you can use it to ensure that\nprocesses are started on specific nodes.\n\n> #### Required mappings {: .warning}\n>\n> When using Guided distribution strategy, you **must** provide the child mappings\n> during the initialization. If the child mappings are not provided, the\n> initialization will **fail**.\n>\n> Example:\n>\n> ```elixir\n> iex> child_spec = %{id: :my_child, start: {MyProcess, :start_link, []}}\n> iex> ProcessHub.start_children(:my_hub, [child_spec], [child_mapping: %{my_child: [:node1, :node2]}])\n> ```","ref":"ProcessHub.Strategy.Distribution.Guided.html"},{"type":"function","title":"ProcessHub.Strategy.Distribution.Guided.handle_children_start/2","doc":"","ref":"ProcessHub.Strategy.Distribution.Guided.html#handle_children_start/2"},{"type":"function","title":"ProcessHub.Strategy.Distribution.Guided.insert_child_mappings/2","doc":"","ref":"ProcessHub.Strategy.Distribution.Guided.html#insert_child_mappings/2"},{"type":"type","title":"ProcessHub.Strategy.Distribution.Guided.t/0","doc":"","ref":"ProcessHub.Strategy.Distribution.Guided.html#t:t/0"},{"type":"protocol","title":"ProcessHub.Strategy.Migration.Base","doc":"The migration strategy protocol provides API functions for migrating child processes.","ref":"ProcessHub.Strategy.Migration.Base.html"},{"type":"function","title":"ProcessHub.Strategy.Migration.Base.handle_migration/5","doc":"Migrates processes from the local to the remote node.\n\nProcess migration happens when a new node joins the `ProcessHub` cluster, and some of the\nlocal processes are moved to the newly connected node. This also requires the processes\nto be terminated on the local node.\n\nThe redundancy strategy will deal with scenarios where processes are not terminated locally\nand are duplicated on the new node.","ref":"ProcessHub.Strategy.Migration.Base.html#handle_migration/5"},{"type":"function","title":"ProcessHub.Strategy.Migration.Base.init/2","doc":"Triggered when coordinator is initialized.\n\nCould be used to perform initialization.","ref":"ProcessHub.Strategy.Migration.Base.html#init/2"},{"type":"type","title":"ProcessHub.Strategy.Migration.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.Migration.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Migration.ColdSwap","doc":"The cold swap migration strategy implements the `ProcessHub.Strategy.Migration.Base` protocol.\nIt provides a migration strategy where the local process is terminated before starting it on\nthe remote node.\n\nCold swap is a safe strategy if we want to ensure that the child process is not\nrunning on multiple nodes at the same time.\n\nThis is the default strategy for process migration.","ref":"ProcessHub.Strategy.Migration.ColdSwap.html"},{"type":"type","title":"ProcessHub.Strategy.Migration.ColdSwap.t/0","doc":"The cold swap migration struct.\n\nThis struct does not contain any configuration options.","ref":"ProcessHub.Strategy.Migration.ColdSwap.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Migration.HotSwap","doc":"The hot swap migration strategy implements the `ProcessHub.Strategy.Migration.Base` protocol.\nIt provides a migration strategy where the local process is terminated after the new one is\nstarted on the remote node.\n\nIn the following text, we will refer to the process that is being terminated after the migration\nas the peer process.\n\nHotswap migration can also handle process state migration when nodes are leaving the cluster.\nThe process states are stored in the local storage and are sent to the remote node before the\nlocal process is terminated. This will only work if the node is being shut down **gracefully**.\n\nThe hot swap strategy is useful when we want to ensure that there is no downtime when migrating\nthe child process to the remote node. It also provides a way to ensure that the state of the\nprocess is synchronized before terminating the peer process.\n\nTo pass the process state to the newly started process on the remote node, the\n`:handover` option must be set to `true` in the migration strategy configuration, and\nnecessary messages must be handled in the processes.\n\n> #### Using the handover option {: .warning}\n>\n> When the `:handover` option is set to `true`, the peer process must handle the following message:\n> `{:process_hub, :handover_start, startup_resp, from}`.\n\nExample migration with handover using `GenServer`:\n\n```elixir\ndef handle_info({:process_hub, :handover_start, startup_resp, child_id, from}, state) do\n  case startup_resp do\n    {:ok, pid} ->\n      # Send the state to the remote process.\n      Process.send(pid, {:process_hub, :handover, state}, [])\n\n      # Signal the handler process that the state handover has been handled.\n      Process.send(from, {:process_hub, :retention_handled, child_id}, [])\n\n    _ ->\n      nil\n  end\n\n  {:noreply, state}\nend\n\ndef handle_info({:process_hub, :handover, handover_state}, _state) do\n  {:noreply, handover_state}\nend\n```\n\n> #### Use HotSwap macro to provide the handover callbacks automatically.  {: .info}\n> It's convenient to use the `HotSwap` macro to provide the handover callbacks automatically.\n>\n> ```elixir\n> use ProcessHub.Strategy.Migration.HotSwap\n> ```","ref":"ProcessHub.Strategy.Migration.HotSwap.html"},{"type":"function","title":"ProcessHub.Strategy.Migration.HotSwap.handle_process_startups/3","doc":"","ref":"ProcessHub.Strategy.Migration.HotSwap.html#handle_process_startups/3"},{"type":"function","title":"ProcessHub.Strategy.Migration.HotSwap.handle_shutdown/2","doc":"","ref":"ProcessHub.Strategy.Migration.HotSwap.html#handle_shutdown/2"},{"type":"type","title":"ProcessHub.Strategy.Migration.HotSwap.t/0","doc":"Hot-swap migration strategy configuration.\nAvailable options:\n- `:retention` - An integer value in milliseconds is used to specify how long\n  the peer process should be kept alive after a new child process has been started on the remote node.\n  This option is used to ensure that the peer process has had enough time to perform any cleanup\n  or state synchronization before the local process is terminated.\n  Keep in mind that process will be terminated before the retention time has passed if the peer process\n  has notified the handler process that the state transfer has been handled.\n  The default value is `5000`.\n\n- `:handover` - A boolean value. If set to `true`, the processes involved in the migration\n  must handle the following message: `{:process_hub, :handover_start, startup_resp, from}`.\n  This message will be sent to the peer process that will be terminated after the migration.\n  The variable `startup_resp` will contain the response from the `ProcessHub.DistributedSupervisor.start_child/2` function, and if\n  successful, it will be `{:ok, pid}`. The PID can be used to send the state of the process to the\n  remote process.\n  If the `retention` option is used, then the peer process has to signal the handler process\n  that the state has been handled; otherwise, the handler will wait until the default retention time has passed.\n  The handler PID is passed in the `from` variable.\n  The default value is `false`.\n\n- `:handover_data_wait` - An integer value in milliseconds is used to specify how long the handler process\n  should wait for the state of the remote process to be sent to the local node.\n  The default value is `3000`.\n\n- `:child_migration_timeout` - An integer value in milliseconds is used to specify the timeout for single\nchild process migration. If the child process migration does not complete within this time, the migration\nfor single child process will be considered failed but the migration for other child processes will continue.\n  The default value is `10000`.","ref":"ProcessHub.Strategy.Migration.HotSwap.html#t:t/0"},{"type":"protocol","title":"ProcessHub.Strategy.PartitionTolerance.Base","doc":"The partition tolerance strategy protocol defines the behavior for handling\nnode up and down events in the `ProcessHub` cluster.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html"},{"type":"function","title":"ProcessHub.Strategy.PartitionTolerance.Base.init/2","doc":"Triggered when coordinator is initialized.\n\nCould be used to perform initialization.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#init/2"},{"type":"function","title":"ProcessHub.Strategy.PartitionTolerance.Base.toggle_lock?/3","doc":"Determines if the lock should be toggled when a node leaves the cluster.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#toggle_lock?/3"},{"type":"function","title":"ProcessHub.Strategy.PartitionTolerance.Base.toggle_unlock?/3","doc":"Determines if the lock should be released when a node joins the cluster.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#toggle_unlock?/3"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.PartitionTolerance.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.PartitionTolerance.Divergence","doc":"The divergence strategy for partition tolerance is used when the `ProcessHub` cluster is not\nconcerned with partition failures.\n\nWhen a node leaves or joins the `ProcessHub` cluster, the divergence strategy keeps the `ProcessHub`\nrunning no matter what.\n\n> #### `In case of network partition` {: .info}\n> The divergence strategy does not handle network partitions in any way. When a network\n> partition occurs, the `ProcessHub` cluster will continue to run on both sides of the partition.\n\nThis is the default strategy for partition tolerance.","ref":"ProcessHub.Strategy.PartitionTolerance.Divergence.html"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.Divergence.t/0","doc":"Divergence strategy configuration.","ref":"ProcessHub.Strategy.PartitionTolerance.Divergence.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum","doc":"The dynamic quorum strategy provides a dynamic way to handle network partitions in the `ProcessHub` cluster.\n\nThis strategy is suitable for clusters that do not know the number of nodes in the cluster beforehand.\n\nThe dynamic quorum strategy uses a cache to store the quorum log. The quorum log is a list of tuples\ncontaining the node name and the timestamp when the connection to the node was lost.\nThis log is used to calculate the quorum size, which is measured in percentage.\nIf the calculated quorum size is less than the `quorum_size` defined in the configuration,\nthe `ProcessHub` will be considered to be in a network partition.\nIn such a case, the `ProcessHub` will terminate its distributed supervisor process,\n`ProcessHub.DistributedSupervisor`, and all its children. Also, the local event queue will be\nlocked by increasing the priority level.\n\nThe quorum size is calculated by the following formula:\n\n    quorum_size = (connected_nodes / (connected_nodes + down_nodes)) * 100\n\n    # The `connected_nodes` are the number of nodes that are currently connected to the `ProcessHub` cluster.\n    # The `down_nodes` are the number of nodes that are listed under the quorum log.\n\nThe `threshold_time` is used to determine how long a node should be listed in the quorum log\nbefore it is removed from the quorum log. This means if we have a `threshold_time` of 10 seconds,\nand we lose one node every 10 seconds until we have lost all nodes, the system won't be\nconsidered to be in a network partition, regardless of whether the lost nodes are lost due to network\npartition or not. If we lose all those nodes in less than 10 seconds, the system will be considered\nto be in a network partition.\n\n> #### `Scaling down` {: .error}\n> When scaling down the cluster, make sure to scale down one node at a time\n> and wait for the `threshold_time` before scaling down the next node.\n>\n> Do not scale down too many nodes at once because the system may think that it is in a network partition\n> and terminate the `ProcessHub.DistributedSupervisor` process and all its children.","ref":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum.html"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum.t/0","doc":"Dynamic quorum strategy configuration.\n\n- `quorum_size` - The quorum size is measured in percentage. This is the required quorum size.\nFor example, `50` means that if we have 10 nodes in the cluster, the system will be\nconsidered to be in a network partition if we lose 6 (60% of the cluster) nodes or more within\nthe `threshold_time` period.\n- `threshold_time` - The threshold time is measured in seconds. This is how long a node should be\nlisted in the quorum log before it is removed from the quorum log.","ref":"ProcessHub.Strategy.PartitionTolerance.DynamicQuorum.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum","doc":"The static quorum strategy for partition tolerance is used when the `ProcessHub` cluster is\nconcerned with partition failures, and the cluster size is known.\n\nWhen a node leaves or joins the `ProcessHub` cluster, the static quorum strategy will check if the\nquorum is present. If the quorum is not present, the `ProcessHub` will be considered in a network\npartition, and the distributed supervisor process of the `ProcessHub` will be terminated along with\nall the child processes.","ref":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum.html"},{"type":"type","title":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum.t/0","doc":"Static quorum strategy configuration.\n\n- `quorum_size` - The quorum size is measured in the number of nodes. For example, `3` means that\n  there should be at least 3 nodes in the cluster for the `ProcessHub` to be considered healthy.\n- `startup_confirm` - If set to `true`, the `ProcessHub` will check if the quorum is present\n  when the `ProcessHub` is starting up. If the quorum is not present, it will be\n  considered as a network partition. The default value is `false`.","ref":"ProcessHub.Strategy.PartitionTolerance.StaticQuorum.html#t:t/0"},{"type":"protocol","title":"ProcessHub.Strategy.Redundancy.Base","doc":"The redundancy protocol relies on the `HashRing` library to distribute processes across\nthe cluster and determine which node should be responsible for a given process by its `child_id` key.\n\nIt is possible to start the same process on multiple nodes in the cluster.","ref":"ProcessHub.Strategy.Redundancy.Base.html"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Base.init/2","doc":"Triggered when coordinator is initialized.\n\nCould be used to perform initialization.","ref":"ProcessHub.Strategy.Redundancy.Base.html#init/2"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Base.master_node/4","doc":"Returns the master node that the given `child_id` belongs to.","ref":"ProcessHub.Strategy.Redundancy.Base.html#master_node/4"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Base.replication_factor/1","doc":"Returns the replication factor for the given strategy struct. This is the number of replicas\nthat the process will be started with.","ref":"ProcessHub.Strategy.Redundancy.Base.html#replication_factor/1"},{"type":"type","title":"ProcessHub.Strategy.Redundancy.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.Redundancy.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Redundancy.Replication","doc":"The replication strategy allows for multiple instances of a process to be started\nacross the cluster. The number of instances is determined by the `replication_factor`\noption.\n\nThe replication strategy also comes with a `replication_model` option, which determines\nwhich instances of the process are active and which are passive.\nImagine a scenario where you have a process that is responsible for handling streams of\ndata. You may want to have multiple instances of this process running across the cluster,\nbut only one of them should be active at any given time and insert data into the database.\nThe passive instances of the process should be ready to take over if the active instance\nfails.\n\n> Replication strategy selects the master/active node based on the `child_id` and\n> `child_nodes` arguments. The `child_id` is converted to a charlist and summed up.\n> The sum is then used to calculate the index of the master node in the `child_nodes` list.\n> This ensures that the same master node is selected for the same `child_id` and `child_nodes`\n> arguments.\n\nThis strategy also allows replication on all nodes in the cluster. This is done by setting\nthe `replication_factor` option to `:cluster_size`.\n\n`ProcessHub` can notify the process when it is active or passive by sending a `:redundancy_signal`.\n\n> #### `Using the :redundancy_signal option` {: .info}\n> When using the `:redundancy_signal` option, make sure that the processes are handling\n> the message `{:process_hub, :redundancy_signal, mode}`, where the `mode` variable is either\n> `:active` or `:passive`.\n\nExample `GenServer` process handling the `:redundancy_signal`:\n    def handle_info({:process_hub, :redundancy_signal, mode}, state) do\n      # Update the state with the new mode and do something differently.\n      {:noreply, Map.put(state, :replication_mode, mode)}\n    end","ref":"ProcessHub.Strategy.Redundancy.Replication.html"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Replication.handle_post_start/3","doc":"","ref":"ProcessHub.Strategy.Redundancy.Replication.html#handle_post_start/3"},{"type":"function","title":"ProcessHub.Strategy.Redundancy.Replication.handle_post_update/3","doc":"","ref":"ProcessHub.Strategy.Redundancy.Replication.html#handle_post_update/3"},{"type":"type","title":"ProcessHub.Strategy.Redundancy.Replication.t/0","doc":"Replication strategy options.\n\n- `replication_factor` - The number of instances of a single process across the cluster.\nThis option can be positive integer or `cluster_size` atom to replicate process on all nodes.\nDefault value is `2`.\n- `replication_model` - This option determines the mode in which the instances of processes are started.\nDefault value is `:active_active`.\n  - `:active_active` - All instances of a process across the cluster are equal.\n  - `:active_passive` - Only one instance of a process across the cluster is active; the rest are passive.\n    Remaining replicas are started as passive processes.\n- `redundancy_signal` - This option determines when a process should be notified of its replication mode.\nDefault value is `:none`.\n  - `:none` - No notifications are sent.\n  - `:active_active` - Only active processes are notified.\n  - `:active_passive` - Only passive processes are notified.\n  - `:all` - All processes are notified.","ref":"ProcessHub.Strategy.Redundancy.Replication.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Redundancy.Singularity","doc":"The Singularity strategy starts a single instance of a process and creates no replicas on\nother nodes. This is the default strategy.","ref":"ProcessHub.Strategy.Redundancy.Singularity.html"},{"type":"type","title":"ProcessHub.Strategy.Redundancy.Singularity.t/0","doc":"No options are available.","ref":"ProcessHub.Strategy.Redundancy.Singularity.html#t:t/0"},{"type":"protocol","title":"ProcessHub.Strategy.Synchronization.Base","doc":"This protocol defines the behavior of a synchronization strategy.","ref":"ProcessHub.Strategy.Synchronization.Base.html"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.handle_synchronization/4","doc":"Handles the periodic synchronization of the process registry.","ref":"ProcessHub.Strategy.Synchronization.Base.html#handle_synchronization/4"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.init/2","doc":"Triggered when coordinator is initialized.\n\nCould be used to perform initialization.","ref":"ProcessHub.Strategy.Synchronization.Base.html#init/2"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.init_sync/3","doc":"Initializes the periodic synchronization of the process registry.","ref":"ProcessHub.Strategy.Synchronization.Base.html#init_sync/3"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Base.propagate/6","doc":"This function is called when a process has been started on the local node, and the\ninformation about the process is about to be propagated to other nodes.","ref":"ProcessHub.Strategy.Synchronization.Base.html#propagate/6"},{"type":"type","title":"ProcessHub.Strategy.Synchronization.Base.t/0","doc":"All the types that implement this protocol.","ref":"ProcessHub.Strategy.Synchronization.Base.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Synchronization.Gossip","doc":"The Gossip synchronization strategy provides a method for spreading information to other nodes\nwithin the `ProcessHub` cluster. It utilizes a gossip protocol to\nsynchronize the process registry across the cluster.\n\nThe Gossip strategy is most suitable for clusters are large. It scales well but produces\nhigher latency than the PubSub strategy when operating in small clusters.\nWhen the cluster increases in size, Gossip protocol can also save bandwidth compared to PubSub.\n\n> The Gossip strategy works as follows:\n> - The synchronization process is initiated on a single node.\n> - The node collects its own local process registry data and appends it to the synchronization data.\n> - It selects a predefined number of nodes that have not yet added their local registry data.\n> - The node sends the data to the selected nodes.\n> - The nodes append their local registry data to the received data and send it to the next nodes.\n> - When all nodes have added their data to the synchronization data, the message will be sent to\n> nodes that have not yet acknowledged the synchronization ack.\n> - If node receives the synchronization data which contains all nodes data, it will\n> synchronize the data with it's local process registry and forward the data to the next nodes\n> that have not yet acknowledged the synchronization ack.\n> - When all nodes in the cluster have acknowledged the synchronization data, the synchronization\n> process is completed and the reference is invalidated.\n\nEach node also adds a timestamp to the synchronization data. This is used to ensure that\nthe synchronization data is not older than the data that is already in the local process registry.","ref":"ProcessHub.Strategy.Synchronization.Gossip.html"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Gossip.handle_propagation/4","doc":"","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#handle_propagation/4"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Gossip.handle_propagation_type/4","doc":"","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#handle_propagation_type/4"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Gossip.invalidate_ref/3","doc":"","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#invalidate_ref/3"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Gossip.propagate_data/5","doc":"","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#propagate_data/5"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Gossip.recipients_select/2","doc":"","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#recipients_select/2"},{"type":"function","title":"ProcessHub.Strategy.Synchronization.Gossip.unacked_nodes/2","doc":"","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#unacked_nodes/2"},{"type":"type","title":"ProcessHub.Strategy.Synchronization.Gossip.t/0","doc":"The Gossip strategy configuration options.\n\n* `sync_interval` - The periodic synchronization interval in milliseconds. The default is `15000`.\n* `recipients` - The number of nodes that will receive the synchronization data and propagate it further. The default is `3`.\n* `restricted_init` - If set to `true`, the synchronization process will only be started on a single node.\n  This node is selected by sorting the node names alphabetically and selecting the first node. The default is `true`.","ref":"ProcessHub.Strategy.Synchronization.Gossip.html#t:t/0"},{"type":"module","title":"ProcessHub.Strategy.Synchronization.PubSub","doc":"This PubSub synchronization strategy uses the `:blockade` library to\ndispatch and handle synchronization events. Each `ProcessHub` instance\nhas its own event queue that is used to dispatch and handle synchronization events.","ref":"ProcessHub.Strategy.Synchronization.PubSub.html"},{"type":"type","title":"ProcessHub.Strategy.Synchronization.PubSub.t/0","doc":"The PubSub synchronization strategy options.\n\n- `sync_interval` - the periodic synchronization interval in milliseconds. Defaults to `15000`.","ref":"ProcessHub.Strategy.Synchronization.PubSub.html#t:t/0"},{"type":"module","title":"ProcessHub.Utility.Bag","doc":"Utility functions.","ref":"ProcessHub.Utility.Bag.html"},{"type":"function","title":"ProcessHub.Utility.Bag.all_messages/0","doc":"Returns all messages in the mailbox.","ref":"ProcessHub.Utility.Bag.html#all_messages/0"},{"type":"function","title":"ProcessHub.Utility.Bag.gen_child_specs/2","doc":"Generates child specs for testing.","ref":"ProcessHub.Utility.Bag.html#gen_child_specs/2"},{"type":"function","title":"ProcessHub.Utility.Bag.hook_erlang_send/3","doc":"Sends hook messages to the given process.","ref":"ProcessHub.Utility.Bag.html#hook_erlang_send/3"},{"type":"function","title":"ProcessHub.Utility.Bag.receive_multiple/3","doc":"Waits and receives multiple messages.","ref":"ProcessHub.Utility.Bag.html#receive_multiple/3"},{"type":"function","title":"ProcessHub.Utility.Bag.timestamp/1","doc":"Returns the current timestamp in the given precision.","ref":"ProcessHub.Utility.Bag.html#timestamp/1"},{"type":"type","title":"ProcessHub.Utility.Bag.gen_child_specs_opts/0","doc":"","ref":"ProcessHub.Utility.Bag.html#t:gen_child_specs_opts/0"},{"type":"module","title":"ProcessHub.Utility.Name","doc":"Utility functions for generating names processes/ETS tables registered under\nthe `hub_id`.","ref":"ProcessHub.Utility.Name.html"},{"type":"function","title":"ProcessHub.Utility.Name.concat_name/2","doc":"Concatenates the given atoms/binary with the given separator.","ref":"ProcessHub.Utility.Name.html#concat_name/2"},{"type":"function","title":"ProcessHub.Utility.Name.coordinator/1","doc":"The name of main coordinator process.","ref":"ProcessHub.Utility.Name.html#coordinator/1"},{"type":"function","title":"ProcessHub.Utility.Name.distributed_supervisor/1","doc":"The name distributed supervisor process.","ref":"ProcessHub.Utility.Name.html#distributed_supervisor/1"},{"type":"function","title":"ProcessHub.Utility.Name.event_queue/1","doc":"The name of the event queue process.","ref":"ProcessHub.Utility.Name.html#event_queue/1"},{"type":"function","title":"ProcessHub.Utility.Name.extract_hub_id/1","doc":"Extracts the `hub_id` from the given name.","ref":"ProcessHub.Utility.Name.html#extract_hub_id/1"},{"type":"function","title":"ProcessHub.Utility.Name.hook_registry/1","doc":"The name of the hook registry process.","ref":"ProcessHub.Utility.Name.html#hook_registry/1"},{"type":"function","title":"ProcessHub.Utility.Name.initializer/1","doc":"The name of the main initializer process.","ref":"ProcessHub.Utility.Name.html#initializer/1"},{"type":"function","title":"ProcessHub.Utility.Name.janitor/1","doc":"The name of the janitor process.","ref":"ProcessHub.Utility.Name.html#janitor/1"},{"type":"function","title":"ProcessHub.Utility.Name.local_storage/1","doc":"Returns the localstorage identifier.","ref":"ProcessHub.Utility.Name.html#local_storage/1"},{"type":"function","title":"ProcessHub.Utility.Name.registry/1","doc":"Returns the process registry table identifier.","ref":"ProcessHub.Utility.Name.html#registry/1"},{"type":"function","title":"ProcessHub.Utility.Name.task_supervisor/1","doc":"The name of the task supervisor process.","ref":"ProcessHub.Utility.Name.html#task_supervisor/1"},{"type":"function","title":"ProcessHub.Utility.Name.worker_queue/1","doc":"","ref":"ProcessHub.Utility.Name.html#worker_queue/1"},{"type":"module","title":"ProcessHub.WorkerQueue","doc":"","ref":"ProcessHub.WorkerQueue.html"},{"type":"function","title":"ProcessHub.WorkerQueue.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"ProcessHub.WorkerQueue.html#child_spec/1"},{"type":"function","title":"ProcessHub.WorkerQueue.start_link/1","doc":"","ref":"ProcessHub.WorkerQueue.html#start_link/1"},{"type":"extras","title":"ProcessHub","doc":"# ProcessHub\n\n![example workflow](https://github.com/alfetahe/process-hub/actions/workflows/elixir.yml/badge.svg)  [![hex.pm version](https://img.shields.io/hexpm/v/coverex.svg?style=flat)](https://hex.pm/packages/process_hub)","ref":"readme.html"},{"type":"extras","title":"Description - ProcessHub","doc":"Library for building distributed systems that are scalable. It handles the distribution of\nprocesses within a cluster of nodes while providing a globally synchronized process registry.\n\nProcessHub takes care of starting, stopping and monitoring processes in the cluster.\nIt scales automatically when cluster is updated and handles network partitions.\n\nBuilding distributed systems is hard and designing one is all about trade-offs.\nThere are many things to consider and each system has its own requirements. \nThis library aims to be flexible and configurable to suit different use cases.\n\nProcessHub is designed to be **decentralized** in its architecture. It does not rely on a\nsingle node to manage the cluster. Each node in the cluster is considered equal.\nThe default distribution strategy is based on consistent hashing.\n\nDetailed documentation can be found at [https://hexdocs.pm/process_hub](https://hexdocs.pm/process_hub).\n\n> ProcessHub is built with scalability and availability in mind.\n> Most of the operations are asynchronous and non-blocking.\n> It can guarantee **eventual** consistency.\n\nProcessHub provides a set of configurable strategies for building distributed\napplications in Elixir.\n\n> #### ProcessHub is eventually consistent {: .info}\n> ProcessHub is built with scalability and availability in mind. \n> Most of the operations are asynchronous and non-blocking. It can guarantee **eventual consistency**.\n>\n> this means that the system may not be in a consistent state at all times, \n> but it will eventually converge to a consistent state.","ref":"readme.html#description"},{"type":"extras","title":"Features - ProcessHub","doc":"Main features include:\n- Distributing processes within a cluster of nodes.\n- Distributed and synchronized process registry for fast process lookups.\n- Process state handover.\n- Strategies to handle network partitions, node failures, process migrations,\nsynchronization, distribution and more.\n- Hooks for triggering events on specific actions and extend the functionality.\n- Automatic hub cluster forming and healing when nodes join or leave the cluster.\n- Ability to define custom strategies to alter the behavior of the system.","ref":"readme.html#features"},{"type":"extras","title":"Installation - ProcessHub","doc":"1. Add `process_hub` to your list of dependencies in `mix.exs`:\n\n    ```elixir\n    def deps do\n      [\n        {:process_hub, \"~> 0.2.7-alpha\"}\n      ]\n    end\n    ```\n\n2. Start the `ProcessHub` supervisor under your application supervision tree:\n\n    ```elixir\n    defmodule MyApp.Application do\n      use Application\n\n      def start(_type, _args) do\n        children = [\n          ProcessHub.child_spec(%ProcessHub{hub_id: :my_hub})\n        ]\n\n        opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n        Supervisor.start_link(children, opts)\n      end\n    end\n    ```\n  It is possible to start multiple hubs under the same supervision tree.\n  Each hub must have a unique `:hub_id`.","ref":"readme.html#installation"},{"type":"extras","title":"Contributing - ProcessHub","doc":"Contributions are welcome and appreciated. If you have any ideas, suggestions, or bugs to report,\nplease open an issue or a pull request on GitHub.","ref":"readme.html#contributing"},{"type":"extras","title":"Getting started","doc":"# Getting started\n\nIn this guide we will go through the basic usage of `ProcessHub` and how to integrate \nit in your application.\n\nTo get more familiar with the public API of `ProcessHub`, please refer to the `ProcessHub` module.","ref":"introduction.html"},{"type":"extras","title":"Installation - Getting started","doc":"In order to use `ProcessHub` in your application, you need to add it as a dependency.\nOnce you have added it as a dependency, you need to start the `ProcessHub` master supervisor.\nIt is highly recommended to start the `ProcessHub` supervisor under your application \nsupervision tree.\n\n1. Add `process_hub` to your list of dependencies in `mix.exs`:\n    ```elixir\n    def deps do\n        [\n            {:process_hub, \"~> 0.2.7-alpha\"}\n        ]\n    end\n    ```\n\n2. Pull the dependency:\n    ```bash\n    mix deps.get\n    ```\n\n3. Start the `ProcessHub` supervisor under your application supervision tree:\n    ```elixir\n    defmodule MyApp.Application do\n        use Application\n\n        def start(_type, _args) do\n            children = [\n                ProcessHub.child_spec(%ProcessHub{hub_id: :my_hub})\n            ]\n\n            opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n            Supervisor.start_link(children, opts)\n        end\n    end\n    ```\n\n> #### Hub ID {: .neutral}\n>\n> Each node in the cluster that starts a `ProcessHub` instance and shares the same `:hub_id` will \n> form a hub cluster.\n>\n> It is possible to start multiple hubs under the same supervision tree with\n> different configuration options, in this case, each hub must have a unique `:hub_id`.","ref":"introduction.html#installation"},{"type":"extras","title":"Example usage - Getting started","doc":"The following example shows how to start 2 elixir nodes, connect them and start processes\nunder the `ProcessHub` cluster. This demonstrates how the processes are distributed within\nthe cluster.\n\n**Note:** The examples below assume that the `ProcessHub` is already started under the\nsupervision tree. If not please refer to the [Installation](#installation) section.\n\n**Note:** Make sure you have a GenServer module called `MyProcess` defined in your project.\n```elixir\ndefmodule MyProcess do\n    use GenServer\n\n    def start_link() do\n        GenServer.start_link(__MODULE__, nil)\n    end\n\n    def init(_) do\n        {:ok, nil}\n    end\nend\n```\n\n<!-- tabs-open -->","ref":"introduction.html#example-usage"},{"type":"extras","title":"Node 1 - Getting started","doc":"Start the first node with the following command:\n\n```bash\niex --name node1@127.0.0.1 --cookie mycookie -S mix\n```\n\n```elixir\n# Run the following in the iex console to start 5 processes under the hub.\niex> ProcessHub.start_children(:my_hub, [\n...>    %{id: :process1, start: {MyProcess, :start_link, []}},\n...>    %{id: :process2, start: {MyProcess, :start_link, []}},\n...>    %{id: :process3, start: {MyProcess, :start_link, []}},\n...>    %{id: :process4, start: {MyProcess, :start_link, []}},\n...>    %{id: :process5, start: {MyProcess, :start_link, []}}\n...> ])\n{:ok, :start_initiated}\n\n# Check the started processes by running the command below.\niex> ProcessHub.process_list(:my_hub, :global)\n[\n  process1: [\"node1@127.0.0.1\": #PID<0.256.0>],\n  process2: [\"node1@127.0.0.1\": #PID<0.257.0>],\n  process3: [\"node1@127.0.0.1\": #PID<0.258.0>],\n  process4: [\"node1@127.0.0.1\": #PID<0.259.0>],\n  process5: [\"node1@127.0.0.1\": #PID<0.260.0>]\n]\n```","ref":"introduction.html#node-1"},{"type":"extras","title":"Node 2 - Getting started","doc":"We will use this node to connect to the first node and see how the processes are\nautomatically distributed.\n\nStart the second node.\n```bash\niex --name node2@127.0.0.1 --cookie mycookie -S mix\n```\n\n```elixir\n# Connect the second node to the first node.\niex> Node.connect(:\"node1@127.0.0.1\")\ntrue\n\n# Check the started procsses by running the command below and\n# see how some of the processes are distributed to the second node.\niex> ProcessHub.process_list(:my_hub, :global)\n[\n  process4: [\"node2@127.0.0.1\": #PID<0.285.0>],\n  process3: [\"node2@127.0.0.1\": #PID<0.284.0>],\n  process1: [\"node2@127.0.0.1\": #PID<0.283.0>],\n  process2: [\"node1@127.0.0.1\": #PID<23045.257.0>],\n  process5: [\"node1@127.0.0.1\": #PID<23045.260.0>]\n]\n```\n<!-- tabs-close -->","ref":"introduction.html#node-2"},{"type":"extras","title":"ProcessRegistry","doc":"# ProcessRegistry\n\nThe `ProcessRegistry` is a global registry that keeps track of all the processes in the system. \nIt is used to look up a process by its name, and to register a process with a name.\n\nIt's recommended to interact with `ProcessRegistry` through the main `ProcessHub` API module.\n\nBy default `ProcessRegistry` is using pubs/subs mechanism to notify subscribers about changes in the registry. This implementation can be replaced with `ProcessHub.Strategy.Synchronization.Gossip` or\ncustom implementation.\n\nWe will skip the boiler plate code for starting the child processes and focus on the registry itself.","ref":"processregistry.html"},{"type":"extras","title":"Querying the whole registry - ProcessRegistry","doc":"We will start of by querying the whole registry with `ProcessHub.process_list/2` function.\n\nThis function allows us to query the whole registry or a specific hub and specify if we want to include remote processes or not.\n\nIf we choose to exclude remote processes, the node names wont be included in the returned list.\nWe can switch between `:global` and `:local` mode by passing the second argument to the function.\n\nThe `ProcessHub.process_list/2`function is the most efficient way to query the whole registry because\nthe data is coming from the local registry and no network calls are made.\n\n\n```elixir\niex> ProcessHub.process_list(:my_hub, :global)\n[\n  my_process_1: [two@anuar: #PID<23772.233.0>],\n  my_process_2: [one@anuar: #PID<0.250.0>],\n  my_process_3: [one@anuar: #PID<0.253.0>],\n  my_process_4: [one@anuar: #PID<0.256.0>],\n  my_process_5: [one@anuar: #PID<0.259.0>],\n  my_process_6: [one@anuar: #PID<0.262.0>],\n  my_process_7: [one@anuar: #PID<0.265.0>],\n  my_process_8: [two@anuar: #PID<23772.254.0>],\n  my_process_9: [one@anuar: #PID<0.271.0>],\n  my_process_10: [one@anuar: #PID<0.274.0>]\n]\n\niex> ProcessHub.process_list(:my_hub, :local)\n[\n  my_process_2: #PID<0.250.0>,\n  my_process_3: #PID<0.253.0>,\n  my_process_4: #PID<0.256.0>,\n  my_process_5: #PID<0.259.0>,\n  my_process_6: #PID<0.262.0>,\n  my_process_7: #PID<0.265.0>,\n  my_process_9: #PID<0.271.0>,\n  my_process_10: #PID<0.274.0>\n]\n```\n\n\nAnother similar option is `ProcessHub.process_registry/1` which dumps the whole registry into a map\nincluding other metadata stored in the registry.\n```elixir\niex> ProcessHub.process_registry(:my_hub)\n%{\n  my_process_1: {%{id: :my_process_1, start: {MyProcess, :start_link, []}},\n   [two@anuar: #PID<23772.233.0>]},\n  my_process_2: {%{id: :my_process_2, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.250.0>]},\n  my_process_3: {%{id: :my_process_3, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.253.0>]},\n  my_process_4: {%{id: :my_process_4, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.256.0>]},\n  my_process_5: {%{id: :my_process_5, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.259.0>]},\n  my_process_6: {%{id: :my_process_6, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.262.0>]},\n  my_process_7: {%{id: :my_process_7, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.265.0>]},\n  my_process_8: {%{id: :my_process_8, start: {MyProcess, :start_link, []}},\n   [two@anuar: #PID<23772.254.0>]},\n  my_process_9: {%{id: :my_process_9, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.271.0>]},\n  my_process_10: {%{id: :my_process_10, start: {MyProcess, :start_link, []}},\n   [one@anuar: #PID<0.274.0>]}\n}\n```\n\n\nThere's also a third option `ProcessHub.which_children/2` but is highly discouraged to use and may \nbe deprecated in the future.\n\nThis function does network calls to all nodes in the cluster to get the list of processes and\nalso may lead to memory and performance issues when used with large nuber of processes.\n\nIt's using the `Supervisor.which_children/1` function under the hood which has stated in the \ndocumentation: **Note that calling this function when supervising a large number of children\n  under low memory conditions can cause an out of memory exception.**\n\nHere is an example of how to use `ProcessHub.which_children/2` function:\n\n```elixir\niex> ProcessHub.which_children(:my_hub, [:global])\n[\n  one@anuar: [\n    {:my_process_10, #PID<0.274.0>, :worker, [MyProcess]},\n    {:my_process_9, #PID<0.271.0>, :worker, [MyProcess]},\n    {:my_process_7, #PID<0.265.0>, :worker, [MyProcess]},\n    {:my_process_6, #PID<0.262.0>, :worker, [MyProcess]},\n    {:my_process_5, #PID<0.259.0>, :worker, [MyProcess]},\n    {:my_process_4, #PID<0.256.0>, :worker, [MyProcess]},\n    {:my_process_3, #PID<0.253.0>, :worker, [MyProcess]},\n    {:my_process_2, #PID<0.250.0>, :worker, [MyProcess]}\n  ],\n  two@anuar: [\n    {:my_process_8, #PID<23772.254.0>, :worker, [MyProcess]},\n    {:my_process_1, #PID<23772.233.0>, :worker, [MyProcess]}\n  ]\n]\n\niex> ProcessHub.which_children(:my_hub, [:local])\n{:one@anuar,\n [\n   {:my_process_10, #PID<0.274.0>, :worker, [MyProcess]},\n   {:my_process_9, #PID<0.271.0>, :worker, [MyProcess]},\n   {:my_process_7, #PID<0.265.0>, :worker, [MyProcess]},\n   {:my_process_6, #PID<0.262.0>, :worker, [MyProcess]},\n   {:my_process_5, #PID<0.259.0>, :worker, [MyProcess]},\n   {:my_process_4, #PID<0.256.0>, :worker, [MyProcess]},\n   {:my_process_3, #PID<0.253.0>, :worker, [MyProcess]},\n   {:my_process_2, #PID<0.250.0>, :worker, [MyProcess]}\n ]}\n```","ref":"processregistry.html#querying-the-whole-registry"},{"type":"extras","title":"Querying a specific child - ProcessRegistry","doc":"To quickly get the PID of the running process by it's `child_id` we can use the\n`ProcessHub.get_pid/2` function which returns the process identifier.\n\n```elixir\niex> ProcessHub.get_pid(:my_hub, :my_process_10)\n#PID<0.228.0>\n```\n\nNow when utilizing replication strategies we should default to another function `ProcessHub.get_pids/2`\nwhich returns a list of PIDs for the given `child_id`.\n\n```elixir\niex> ProcessHub.get_pids(:my_hub, :my_process_10)\n[#PID<0.228.0>]\n```\n\nWe may also query the whole child info by using the `ProcessHub.which_children/2` function.\nThis function also gives us the `child_spec`.\n\n```elixir\niex> ProcessHub.child_lookup(:my_hub, :my_process_10)\n{\n    %{id: :my_process_10, start: {MyProcess, :start_link, []}},\n    [one@anuar: #PID<0.228.0>]\n}\n ```","ref":"processregistry.html#querying-a-specific-child"},{"type":"extras","title":"Architecture","doc":"# Architecture","ref":"architecture.html"},{"type":"extras","title":"Table of Contents - Architecture","doc":"* [Overview](#overview)\n* [Cluster Discovery and Formation](#cluster-discovery-and-formation)\n* [Resilience and Reliability](#resilience-and-reliability)\n* [Locking Mechanism](#locking-mechanism)","ref":"architecture.html#table-of-contents"},{"type":"extras","title":"Overview - Architecture","doc":"The overall architecture of ProcessHub is designed to provide a reliable and resilient\ndistributed system that can handle network partitions, node failures, process migrations,\n\nThis is mostly achieved by asynchronous and non-blocking operations, and by using the\n`Supervisor` behavior to monitor and restart processes when they die unexpectedly.\n\nProcessHub is eventually consistent meaning that it can guarantee that the state of the\nsystem will eventually converge to a consistent state. This enables the system to be\nscalable and highly available. \n\nProcessHub internally uses a event-driven communication and ability to define listeners/hooks.\nThis enables some type of way to react to specific events such as process registration etc.\n\nMost of the operations are carried out using special Task processes. These processes are\nstarted on demand by the `ProcessHub.Coordinator` process and are supervised by the `Task.Supervisor`.\n\nThe coordinator process is responsible for coordinating the operations and making sure that\nthe operations are carried out in the correct order. This process is the heart of the system and\nis responsible for the overall functionality of the system.","ref":"architecture.html#overview"},{"type":"extras","title":"Cluster Discovery and Formation - Architecture","doc":"ProcessHub monitors connecting and disconnecting nodes and forms a cluster automatically\nfrom the connected nodes that share the same `hub_id`. It's not required to start\nthe `ProcessHub` on all nodes in the cluster.","ref":"architecture.html#cluster-discovery-and-formation"},{"type":"extras","title":"Resilience and Reliability - Architecture","doc":"ProcessHub uses the `Supervisor` behavior and leverages the features that come with it.\nEach hub starts its own `ProcessHub.DistributedSupervisor` process, which is responsible for\nstarting, stopping, and monitoring the processes in its local cluster.\n\nWhen a process dies unexpectedly, the `ProcessHub.DistributedSupervisor` will restart it\nautomatically.\n\nProcessHub also takes care of validating the `child_spec` before starting it and makes sure\nit's started on the right node that the process belongs to.\nIf the process is being started on the wrong node, the initialization request will be forwarded\nto the correct node.","ref":"architecture.html#resilience-and-reliability"},{"type":"extras","title":"Locking Mechanism - Architecture","doc":"ProcessHub utilizes the `:blockade` library to provide event-driven communication\nand a locking mechanism.\nIt locks the local event queue by increasing its priority for some operations.\nThis allows the system to queue events and process them in order to preserve data integrity.\nOther events can be processed once the priority level is set back to default.\n\nTo avoid deadlocks, the system places a timeout on the event queue priority and\nrestores it to its original value if the timeout is reached.","ref":"architecture.html#locking-mechanism"},{"type":"extras","title":"Configuring the hub","doc":"# Configuring the hub","ref":"configuration.html"},{"type":"extras","title":"Configurable strategies - Configuring the hub","doc":"ProcessHub comes with 11 different strategies that can be used to configure the hub.\nAll strategies are Elixir structs that implement their own base protocol.\n\nIn fact, you can define your own strategies by implementing the base protocols.\n\nWhen configuring the hub, you can pass the strategies as part of the `%ProcessHub{}` struct.\n\nLook at the documentation for each strategy for more information on how to configure them.\n\nAn example can be seen below.\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [process_hub()]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  defp process_hub() do\n    {ProcessHub, %ProcessHub{\n      hub_id: :my_hub,\n      # Configure the redundancy strategy.\n      redundancy_strategy: %ProcessHub.Strategy.Redundancy.Replication{\n        replication_factor: 2,\n        replication_model: :active_passive,\n        redundancy_signal: :none\n      },\n      # Configure the migration strategy.\n      migration_strategy: %ProcessHub.Strategy.Migration.HotSwap{\n        retention: 2000,\n        handover: true\n      },\n      # Configure the synchronization strategy.\n      synchronization_strategy: %ProcessHub.Strategy.Synchronization.PubSub{\n        sync_interval: 10000\n      },\n      # Configure the partition tolerance strategy.\n      partition_tolerance_strategy: %ProcessHub.Strategy.PartitionTolerance.StaticQuorum{\n        quorum_size: 3\n      },\n      # Configure the distribution strategy.\n      distribution_strategy: %ProcessHub.Strategy.Distribution.ConsistentHashing{}\n    }}\n  end\nend\n```\n\nEach strategy module has additional documentation on how to configure it.","ref":"configuration.html#configurable-strategies"},{"type":"extras","title":"Redundancy Strategy - Configuring the hub","doc":"`ProcessHub.Strategy.Redundancy.Base` - defines the base protocol for redundancy strategies.\nThis strategy is used to define how many replicas of a process should be started\nacross the cluster. Starting multiple instances of a process across the cluster is useful\nfor redundancy and fault tolerance.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.Redundancy.Singularity` - only 1 process per `child_id` without\nany replicas. This is also the default strategy and contains no special configuration options.\n- `ProcessHub.Strategy.Redundancy.Replication` - starts multiple replicas of a process\nacross the cluster. The number of replicas is defined by the `:replication_factor`\noption. This strategy also supports `:active_active` and `:active_passive` modes.\nMeaning we may have one active process and the rest are passive.\nThe mode is defined by the `:replication_model` option.\nThis information will be passed to the started process.\nThe default mode is `:active_active`, meaning all processes are equal and considered active.","ref":"configuration.html#redundancy-strategy"},{"type":"extras","title":"Migration Strategy - Configuring the hub","doc":"`ProcessHub.Strategy.Migration.Base` - defines the base protocol for migration strategies.\nThis strategy is used to define how the processes are migrated when a node joins or leaves the cluster.\n\nMigration is the process of moving processes from one node to another.\nOne of the reasons why migration happens is when a node leaves the cluster.\nWhen a node leaves the cluster, it is possible that some processes are still running on that\nnode, so these need to be migrated to another node. Also, when a new node joins the\ncluster, other nodes may migrate some processes over to the new node.\n\nAt the moment, there are 2 migration strategies available:\n- `ProcessHub.Strategy.Migration.ColdSwap` - migrate processes by stopping the process\non the old node before starting it on the new node. This is the default strategy and defines no special configuration options.\n- `ProcessHub.Strategy.Migration.HotSwap` - this strategy is used to migrate processes\nby starting the process on the new node before stopping it on the old node.\nThis strategy is useful when we want to avoid any downtime. This strategy is also\nuseful when the process is stateful, and we want to avoid any data loss by handing over\nthe state from the old process to the new process. See the module for handover examples.","ref":"configuration.html#migration-strategy"},{"type":"extras","title":"Synchronization Strategy - Configuring the hub","doc":"`ProcessHub.Strategy.Synchronization.Base` - defines the base protocol for synchronization\nstrategies which define the method that is used to synchronize the process registry.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.Synchronization.PubSub` - uses a publish/subscribe model to synchronize\nthe process registry. Each node in the cluster will subscribe to a topic and publish any changes to the topic.\nThese changes could be events such as adding or removing a process from the registry.\nThis is the default and recommended synchronization strategy for most users.\n- `ProcessHub.Strategy.Synchronization.Gossip` - uses a gossip protocol to synchronize the\nprocess registry. Using this strategy is recommended when the number of nodes in the cluster\nis large. The Gossip strategy selects a predefined number of nodes to gossip with and\nexchange information about the process registry.\nThese selected nodes will choose other nodes to gossip with and so on until all nodes in the\ncluster are synchronized. It has higher latency than the PubSub strategy specially when the cluster\nis rather small.","ref":"configuration.html#synchronization-strategy"},{"type":"extras","title":"Partition Tolerance Strategy - Configuring the hub","doc":"`ProcessHub.Strategy.PartitionTolerance.Base` - defines the base protocol for partition tolerance\nstrategies which define the method that is used to handle network partitions.\n\nAvailable strategies are:\n- `ProcessHub.Strategy.PartitionTolerance.Divergence` - this strategy is used to handle network\npartitions by diverging the cluster into multiple subclusters. Each subcluster will have its\nown hub and will be considered as a separate cluster. This strategy is the default strategy.\nWhen the network partition is healed, the subclusters will merge back into a single cluster.\n- `ProcessHub.Strategy.PartitionTolerance.StaticQuorum` - this strategy is used to handle network\npartitions by using a static quorum. The quorum size is defined by the `:quorum_size` option.\nWhen a partition happens, the `ProcessHub.DistributedSupervisor` process will terminate\nalong with its children. This strategy is useful when the number of nodes in the cluster is\nknown and rather fixed.\n- `ProcessHub.Strategy.PartitionTolerance.DynamicQuorum` - this strategy is used to handle\nnetwork partitions by using a dynamic quorum. The quorum size is defined by the `:quorum_size`\noption and `:threshold_time` option. The system automatically over time adapts to the number of\nnodes in the cluster. When a partition happens, the `ProcessHub.DistributedSupervisor` process\nwill terminate along with its children.\n  > #### Using DynamicQuorum Strategy {: .error}\n  > When scaling down too many nodes at once, the system may consider itself to be in a\n  > network partition. Read the documentation for the\n  >`ProcessHub.Strategy.PartitionTolerance.DynamicQuorum` strategy for more information.","ref":"configuration.html#partition-tolerance-strategy"},{"type":"extras","title":"Distribution Strategy: - Configuring the hub","doc":"`ProcessHub.Strategy.Distribution.Base` - defines the base protocol for distribution\nstrategies which define the method to determine the node/process mapping.\n\nProcessHub currently supports 2 distribution strategies:\n- `ProcessHub.Strategy.Distribution.ConsistentHashing` - uses a consistent hash ring to determine\nthe node/process mapping. This strategy is the default strategy.\n- `ProcessHub.Strategy.Distribution.Guided` - uses a predefined mapping to determine the\nnode/process mapping. This strategy is useful when we want to control the mapping of\nprocesses to nodes.","ref":"configuration.html#distribution-strategy"},{"type":"extras","title":"Process state handover","doc":"# Process state handover\n\nIn this guide we will show a simple example of how to handover the state of a process from one node to another.\n\nThis scenario is happening automatically when the cluster has received new node event,\nmeaning that the new node has joined the cluster or left.\n\nSuppose we have the following `GenServer` process:\n\n```elixir\ndefmodule MyProcess do\n  use GenServer\n\n  def start_link(state) do\n    GenServer.start_link(__MODULE__, state)\n  end\n\n  def init(state) do\n    {:ok, state}\n  end\nend\n```\n\nIn order to handover the state of the process from one node to another, we need to do the following:","ref":"statehandover.html"},{"type":"extras","title":"1. Configure the hub - Process state handover","doc":"Configure the `ProcessHub` to use the `Hotswap` strategy.\n\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [process_hub()]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  defp process_hub() do\n    {ProcessHub, %ProcessHub{\n      hub_id: :my_hub,\n      migration_strategy: %ProcessHub.Strategy.Migration.HotSwap{\n        retention: 3000,\n        handover: true\n      },\n    }}\n  end\nend\n```\n\nPay attention to the `retention` and `handover` options. The `retention` option is the time in milliseconds that the process state will be kept in the hub after the process has been stopped. The `handover` option is a boolean that indicates whether the process state should be handed over to another node when the process is going to be stopped.\n\n\n> #### Retention option {: .info}\n> The `retention` option does not mean that the old process will be kept that long alive.\n> It is used to limit the time the old processes can stay alive in the handover process\n> before they are killed.","ref":"statehandover.html#1-configure-the-hub"},{"type":"extras","title":"2. Implement the neccessary callbacks - Process state handover","doc":"In order to handover the state our `MyProcess` have to implement the `ProcessHub.Strategy.Migration.Handover` behaviour or define the necessary callbacks.\n\nThe most convenient way is to use the `ProcessHub.Strategy.Migration.HotSwap` module that provides the default implementation of the `ProcessHub.Strategy.Migration.Handover` behaviour.\n\n```elixir\ndefmodule MyProcess do\n  use GenServer\n  use ProcessHub.Strategy.Migration.HotSwap # This has been added\n\n  # ...\nend\n```","ref":"statehandover.html#2-implement-the-neccessary-callbacks"},{"type":"extras","title":"3. That's it! - Process state handover","doc":"Now the `ProcessHub` will take care of the rest. When the process is going to be\nredirected to another node, the state of the process will be handed over to the new process on\nthe new node.","ref":"statehandover.html#3-that-s-it"},{"type":"extras","title":"Handover on node leave - Process state handover","doc":"The hub takes care of the handover process when a node joins the cluster and processes\nmigrate to the new node. \nThe same can be said when a node leaves the cluster, meaning that the process states on the \nleaving node will be handed over to the remaining nodes but only if the node is shut down **gracefully**.\nThis leaves the hub with the time to handover the process states to the remaining nodes.\nIt is advised to scale nodes down one at a time to avoid any data loss(30 second - 10 minute intervals).","ref":"statehandover.html#handover-on-node-leave"},{"type":"extras","title":"More info can be found in the `Hotswap` module - Process state handover","doc":"Check out the [Hotswap](https://hexdocs.pm/process_hub/ProcessHub.Strategy.Migration.HotSwap.html) module for more information.","ref":"statehandover.html#more-info-can-be-found-in-the-hotswap-module"},{"type":"extras","title":"Manual distribution","doc":"# Manual distribution","ref":"manualdistribution.html"},{"type":"extras","title":"Introduction - Manual distribution","doc":"By default `ProcessHub` uses the `ProcessHub.Strategy.Distribution.ConsistentHashing` strategy to distribute processes.\nThis is automatic and the `ProcessHub` will decide which node to start the process on.\nThis strategy is useful for most use cases, but sometimes you may want to distribute processes manually. This guide will show you how to do that.\n\nBy manually distributing processes, you can control which processes are started on which nodes. \nThis can be useful when you want to distribute processes based on specific criteria, such as the node's hardware capabilities or the process's workload.\n\nLet's dive in a see some examples.","ref":"manualdistribution.html#introduction"},{"type":"extras","title":"Example GenServer - Manual distribution","doc":"As always our processes should have some template to follow. Here is a simple `GenServer` process that we will use in our examples:\n\n```elixir\ndefmodule MyProcess do\n  use GenServer\n\n  def start_link(state) do\n    GenServer.start_link(__MODULE__, state)\n  end\n\n  def init(state) do\n    {:ok, state}\n  end\nend\n```","ref":"manualdistribution.html#example-genserver"},{"type":"extras","title":"The configuration - Manual distribution","doc":"First, let's configure the `ProcessHub` to use the `ProcessHub.Strategy.Distribution.Guided` strategy.\n\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [process_hub()]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  defp process_hub() do\n    {ProcessHub, %ProcessHub{\n      hub_id: :my_hub,\n      # We have to swap the default distribution strategy with the guided one.\n      distribution_strategy: %ProcessHub.Strategy.Distribution.Guided{}\n    }}\n  end\nend\n```","ref":"manualdistribution.html#the-configuration"},{"type":"extras","title":"Start 2 nodes - Manual distribution","doc":"Now let's start 2 nodes and connect them to each other.\n\n```bash\niex --name node1@127.0.0.1 --cookie mycookie -S mix\n```\n\n```bash\niex --name node2@127.0.0.1 --cookie mycookie -S mix\n```\n\nFrom node 1, connect to node 2(or vice versa):\n```elixir\nNode.connect(:\"node2@127.0.0.1\")\n```","ref":"manualdistribution.html#start-2-nodes"},{"type":"extras","title":"Start distributed processes. - Manual distribution","doc":"Now let's start a process on node 1 and guide it to node 2.\n\n```elixir\niex> node1_child = %{id: :node1_child, start: {MyProcess, :start_link, [nil]}}\niex> node2_child = %{id: :node2_child, start: {MyProcess, :start_link, [nil]}}\niex> ProcessHub.start_children(:my_hub, [node1_child, node2_child], [\n...>    child_mapping: %{\n...>        node1_child: [:\"node1@127.0.0.1\"],\n...>        node2_child: [:\"node2@127.0.0.1\"],\n...>    }\n...> ])\n```\n\nIn the example above, we started 2 processes, `node1_child` and `node2_child`. We also specified that `node1_child` should be started on node 1 and `node2_child` should be started on node 2.\n\nWe do this by adding the `child_mapping` option to the `ProcessHub.start_children/3` function. The `child_mapping` option is a map that specifies which node to start the process on.\nEach key in the map is the `id` of the process and the value is a list of nodes to start the process on.\n\n> #### Why use lists to specify the node?\n> The reason we use lists for the `child_mapping` is that we can specify multiple nodes for a \n> process if we want to replicate the process on multiple nodes. \n>\n>This requires the `ProcessHub.Strategy.Redundancy.Replication` strategy to be used. Which we will cover in another guide.\n> For example: `node1_child: [:\"node1@127.0.0.1\", :\"node2@127.0.0.1\"]`","ref":"manualdistribution.html#start-distributed-processes"},{"type":"extras","title":"Checking the registered processes - Manual distribution","doc":"Now let's check the registered processes on each node.\n\n```elixir\niex> ProcessHub.process_list(:my_hub, :global)\n[\n  node1_child: [\"node1@127.0.0.1\": #PID<0.280.0>],\n  node2_child: [\"node2@127.0.0.1\": #PID<22474.276.0>]\n]\n```\n\nWe can see that the `node1_child` process is registered on node 1 and the `node2_child` process is registered on node 2.\n\n\n> #### The `child_mapping` is mandatory {: .error}\n> If we decide to use the `ProcessHub.Strategy.Distribution.Guided` strategy, we have to specify the `child_mapping` option when starting the processes. Otherwise `{:error, :missing_child_mapping}` will be returned when starting the processes.","ref":"manualdistribution.html#checking-the-registered-processes"},{"type":"extras","title":"Replicating processes","doc":"# Replicating processes\n\nIn a distributed system, it can be important to replicate processes within a cluster of nodes. This can lead to better resilience and reliability of the system.\nWhen one node unexpectedly fails, the replicated processes on other nodes can take over the work of the failed node immediately.\n\nIn this guide, we will show a simple example of how to replicate processes and prioritize them.","ref":"replicatingprocesses.html"},{"type":"extras","title":"Prerequisites - Replicating processes","doc":"As always we should have some process template to follow. You can re use the `MyProcess` GenServer from the [Manual Distribution](ManualDistribution.md#example-genserver) guide.","ref":"replicatingprocesses.html#prerequisites"},{"type":"extras","title":"The configuration - Replicating processes","doc":"First, let's configure the `ProcessHub` to use the `ProcessHub.Strategy.Redundancy.Replication` strategy.\n\n```elixir\ndefmodule MyApp.Application do\n  use Application\n\n  def start(_type, _args) do\n    children = [process_hub()]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  defp process_hub() do\n    {ProcessHub, %ProcessHub{\n      hub_id: :my_hub,\n      redundancy_strategy: %ProcessHub.Strategy.Redundancy.Replication{    \n        replication_factor: 2,\n        replication_model: :active_passive,\n        redundancy_signal: :all\n      },\n    }}\n  end\nend\n```\n\nLet's explain the used options:\n- `replication_factor`: The number of replicas that will be created for each process. Meaning if the value is 2, then the process will be replicated on 2 nodes.\n- `replication_model`: This is optional and can be either `:active_passive` or `:active_active`. The `:active_passive` model means that only one of the replicas will be marked as active and the others will be passive. We will get into more details about this later.\n- `redundancy_signal`: This is optional and can be used to toggle on/off the message sent\nto the replicas to notify them about their current state (`:passive` or `:active`). This all will be explained in the next sections.","ref":"replicatingprocesses.html#the-configuration"},{"type":"extras","title":"Start 2 nodes - Replicating processes","doc":"Now let's start two nodes and connect them to each other. \nYou can refer to the [Manual Distribution](ManualDistribution.md#start-2-nodes) guide for an example.","ref":"replicatingprocesses.html#start-2-nodes"},{"type":"extras","title":"Start child - Replicating processes","doc":"Next let's start a process on one node and then see how it is replicated on the other node.\n\nWe will be using the `async_wait` just for demonstration purposes and see the results printed in the console.\n\n```elixir\niex> child_spec = %{id: :child1, start: {MyProcess, :start_link, [nil]}}\niex> ProcessHub.start_children(:my_hub, [child_spec], [async_wait: true]) |> ProcessHub.await()\n{:ok,\n [\n   child1: [\n     \"node1@127.0.0.1\": #PID<0.264.0>,\n     \"node2@127.0.0.1\": #PID<22914.996.0>\n   ]\n ]}\n***17:08:08.277 [error] MyProcess #PID<0.264.0> received unexpected message in handle_info/2: {:process_hub, :redundancy_signal, :passive}***\n```\n\nIf we examine the result, we can see that the process `child1` was started on both nodes, meaning it was replicated.\n\n**But what's up with the error message?**\n\nWell, the error message is expected. It is a signal sent to the process to notify it that it is either in passive or active mode. This is part of the `redundancy_signal` option we set in the `ProcessHub` configuration.\n\nWe could have used the `redundancy_signal: :none` option to disable the signal, but we wanted to show you how it works.\n\n> #### Redundancy Signal {: .warning}\n> The `redundancy_signal` option is used to toggle on/off the message sent to the replicas to notify them about their current state (`:passive` or `:active`). \n> This message needs to be handled by the process.\n\nHere's an example how we handle the signal in the `MyProcess`. In the following example, we\nsave the signal in the state and handle incoming messages based on the state.\n\n```elixir\n# The required callback we have to implement, to handle the signal.\ndef handle_info({:process_hub, :redundancy_signal, mode}, state) do\n    # Save the mode in the state for later use.\n    {:noreply, %{mode: mode}}\nend\n\n# Below are examples of how we can handle the signal in the process.\n# These are just examples and you can implement your own logic based on the mode.\n\n# Here we define two different handle_info functions based on the mode.\n# The second one handles the request because we're the active process.\n# The first one does nothing because we're the passive process.\n\ndef handle_info({:handle_request, _request}, %{mode: :passive} = state) do\n    # Ignore and don't do anything because we're the passive process.\n    {:noreply, state}\nend\n\ndef handle_info({:handle_request, request}, %{mode: :aactive} = state) do\n    # Handle the request because we're the active process.\n    handle_request(request)\n    {:noreply, state}\nend\n```","ref":"replicatingprocesses.html#start-child"},{"type":"extras","title":"Conclusion - Replicating processes","doc":"The `ProcessHub.Strategy.Redundancy.Replication` can be used not only for replicating processes but also to manipulate the behavior of the replicas. This can be done by using the `:redundancy_signal` with `:replication_model` options.\n\n`ProcessHub` defines currently two redundancy strategies, You can check these modules for custom configuration options:\n- `ProcessHub.Strategy.Redundancy.Replication`\n- `ProcessHub.Strategy.Redundancy.Singularity`","ref":"replicatingprocesses.html#conclusion"},{"type":"extras","title":"Hooks","doc":"# Hooks\n\nHooks are used to trigger events on specific actions. Hooks can be registered by passing the\nhandlers to the `:hooks` option of the `t:ProcessHub.t/0` configuration struct or by inserting them\ndynamically using the `ProcessHub.Service.HookManager` module.\n\nProcessHub heavily uses hooks internally in the integration tests and strategy implementations.\n\nHooks have to be in the format of an `mfa` tuple. Basically, they are functions that will be called\nwhen the hook is triggered.\n\nIt is possible to register a hook handler with a wildcard argument `:_`, which will be replaced with the hook data when the hook is dispatched.","ref":"hooks.html"},{"type":"extras","title":"Hook registration examples - Hooks","doc":"Example dynamic hook registration using the `t:ProcessHub.t/0` configuration struct:\n```elixir\n# Register a hook handler for the `:pre_cluster_join_hook` event with a wildcard argument.\ndefmodule MyApp.Application do\n  alias ProcessHub.Service.HookManager\n  alias alias ProcessHub.Constant.Hook\n\n  use Application\n\n  def start(_type, _args) do\n    children = [\n      ProcessHub.child_spec(%ProcessHub{\n        hub_id: :my_hub,\n        hooks: %{\n          Hook.pre_cluster_join() => [\n            %HookManager{\n              id: :hook_id_1,\n              m: MyModule,\n              f: :my_function,\n              a: [:some_data, :_]\n            },\n            %HookManager{\n              id: :hook_id_2,\n              m: MyModule2,\n              f: :my_function2,\n              a: [:some_data]\n            }\n          ],\n          Hook.post_cluster_join() => [\n            %HookManager{\n              id: :hook_id_3,\n              m: MyModule,\n              f: :my_function,\n              a: [:some_data, :_]\n            }\n          ]\n        }\n      })\n    ]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\nend\n```\n\nExample hook registration using the `ProcessHub.Service.HookManager` module:\n```elixir\nalias ProcessHub.Service.HookManager\nalias ProcessHub.Constant.Hook\n\n# Register a hook handler for the `:cluster_join` event with a wildcard argument.\nHookManager.register_handler(\n  :my_hub, \n  Hook.post_cluster_join(),\n  %HookManager{\n    id: :hook_id_30, \n    m: MyModule,\n    f: :my_function,\n    a: [:some_data, :_]\n  }\n)\n```","ref":"hooks.html#hook-registration-examples"},{"type":"extras","title":"Hook handler examples - Hooks","doc":"```elixir\n# The hook handler should be in the following format:\ndefmodule MyModule do\n  def my_function(:some_data, dynamic_hook_data) do\n    # Do something with the data.\n  end\nend\n```","ref":"hooks.html#hook-handler-examples"},{"type":"extras","title":"Using hooks to react and handle asynchronous events - Hooks","doc":"Sometimes when starting a child process, it is necessary to know when the child process has been\nstarted and is ready to receive messages. We sure could use the `:async_wait` option but that would\nblock the process that is starting the child process. Instead, we can use hooks to react to the\n`:pre_children_start` event asynchronously.\n\nExample:\n```elixir \n# Register hook handler\nalias ProcessHub.Service.HookManager\nalias ProcessHub.Constant.Hook\n...\nhub_id = :my_hub\n\nhook_handler1 = %HookManager{\n  id: :handler_id_1,\n  m: MyModule,\n  f: :my_first_func,\n  a: [:_]\n}\n\nhook_handler2 = %HookManager{\n  id: :handler_id_2,\n  m: MyModule,\n  f: :my_second_func,\n  a: [:hub_id, :_]\n}\n\n# We're triggering the hook when child process has been registered under the process registry.\nHookManager.register_handlers(:my_hub, Hook.registry_pid_inserted(), [hook_handler1, hook_handler2])\n...\n\n# The hook handlers\ndefmodule MyModule do\n  def my_first_func(hub_id, {child_id, node_pids}) do\n    # Send message to all started child processes.\n    Enum.each(node_pids, fn {_node, pid} ->\n      Process.send(pid, :hello_world, [])\n    end)\n  end\n\n  def my_second_func(hub_id, hook_data) do\n    ...\n  end\nend\n```","ref":"hooks.html#using-hooks-to-react-and-handle-asynchronous-events"},{"type":"extras","title":"Internally registered hook handlers - Hooks","doc":"ProcessHub is using hooks internally to depending on the active configuration.\nThese hook handlers should never be removed or modified.\nThe list of internally registered hooks ids are: \n- `:ch_join`\n- `:ch_leave`\n- `:ch_shutdown`\n- `:mhs_shutdown`\n- `:mhs_process_startups`\n- `:rr_post_start`\n- `:rr_post_stop`\n- `:rr_post_update`\n- `:dg_pre_start_handler`","ref":"hooks.html#internally-registered-hook-handlers"},{"type":"extras","title":"Hooks in custom strategy implementations - Hooks","doc":"Hooks can be used in custom strategy implementations to trigger events on specific actions and\nextend the functionality of the strategy.\n\nHere's an example of ProcessHub using hooks internally in the \n`ProcessHub.Strategy.Distribution.ConsistentHashing` strategy implementation:\n```elixir\nalias ProcessHub.Constant.StorageKey\nalias ProcessHub.Service.Storage\nalias ProcessHub.Service.HookManager\n\n# Register hook handlers to update the hash ring when node joins or leaves the cluster.\ndef init(_strategy, hub_id) do\n  local_storage = ProcessHub.Utility.Name.local_storage(hub_id)\n\n  hub_nodes = Storage.get(local_storage, :hub_nodes)\n\n  Storage.insert(local_storage, StorageKey.hr, Ring.create_ring(hub_nodes))\n\n  join_handler = %HookManager{\n    id: :ch_join,\n    m: ProcessHub.Strategy.Distribution.ConsistentHashing,\n    f: :handle_node_join,\n    a: [hub_id, :_]\n  }\n\n  leave_handler = %HookManager{\n    id: :ch_join,\n    m: ProcessHub.Strategy.Distribution.ConsistentHashing,\n    f: :handle_node_leave,\n    a: [hub_id, :_]\n  }\n\n  shutdown_handler = %HookManager{\n    id: :ch_shutdown,\n    m: ProcessHub.Strategy.Distribution.ConsistentHashing,\n    f: :handle_shutdown,\n    a: [hub_id],\n    p: 100\n  }\n\n  HookManager.register_handler(hub_id, Hook.pre_cluster_join(), join_handler)\n  HookManager.register_handler(hub_id, Hook.pre_cluster_leave(), leave_handler)\n  HookManager.register_handler(hub_id, Hook.coordinator_shutdown(), shutdown_handler)\nend\n\n\n# Handler functions that will be called when the hooks are triggered.\ndef handle_node_join(hub_id, node) do\n  hash_ring = Ring.get_ring(hub_id) |> Ring.add_node(node)\n\n  Storage.insert(ProcessHub.Utility.Name.local_storage(hub_id), StorageKey.hr, hash_ring)\nend\n\ndef handle_node_leave(hub_id, node) do\n  hash_ring = Ring.get_ring(hub_id) |> Ring.remove_node(node)\n\n  Storage.insert(ProcessHub.Utility.Name.local_storage(hub_id), StorageKey.hr, hash_ring)\nend\n```","ref":"hooks.html#hooks-in-custom-strategy-implementations"},{"type":"extras","title":"Hook handler priority - Hooks","doc":"ProcessHub by default does not guarantee the order of hook handler execution unless \nthe hook handler specifies a **priority level**. The priority level is an integer value that\ndetermines the order of execution of the hook handlers. The higher the priority level, the\nearlier the hook handler will be executed. By default all hooks have a priority level of `0`.\n\nWe define the priority level of a hook handler by setting the `:p` field of the `t:ProcessHub.Service.HookManager.t/0`\n\nHere an example of how to set the priority level of a hook handler to 15:\n```elixir\nalias ProcessHub.Service.HookManager\nalias ProcessHub.Constant.Hook\n...\nhook_handler = %HookManager{\n  id: :handler_id_1,\n  m: MyModule,\n  f: :handle_child_start,\n  a: [],\n  p: 15 # here we set the priority level to 15.\n}\n\nHookManager.register_handler(:my_hub, Hook.pre_cluster_leave(), hook_handler)\n```","ref":"hooks.html#hook-handler-priority"},{"type":"extras","title":"Hook handler arguments - Hooks","doc":"The hook handler expects a arguments which is a list of values. \nThese values will then be passed to the hook handler function in the same order.\nSome hooks dispatchers will pass additional data to the hook handler function which\nthen can be accessed using the wildcard argument in the handler function.\nThese are called wildcard arguments and are represented by the `:_` symbol.\n\nExample:\n```elixir\nalias ProcessHub.Service.HookManager\nalias ProcessHub.Constant.Hook\n...\nhook_handler = %HookManager{\n  id: :handler_id_1,\n  m: MyModule,\n  f: :handle_child_start,\n  a: [:_, :some, :something_else] # here the first argument will be the wildcard argument.\n}\n\nHookManager.register_handler(:my_hub, Hook.pre_cluster_join(), hook_handler)\n\n# Here the handler function receives the wildcard argument as the first argument.\ndefmodule MyModule do\n  def handle_child_start(wild_card_value, :some, :something_else) do\n    # Because we used the `:pre_cluster_join_hook` hook event the wild_card_value will be a `node()`\n  end\nend\n```\n\nYou can familiarize yourself with the available hooks and their data in the last section of this guide.","ref":"hooks.html#hook-handler-arguments"},{"type":"extras","title":"Listing hook handlers - Hooks","doc":"You can list all registered hook handlers for a specific hook event by calling \nthe `registered_handlers/2` function from the `ProcessHub.Service.HookManager` module.\n\nExample:\n```elixir\niex> alias ProcessHub.Service.HookManager\niex> alias ProcessHub.Constant.Hook \niex> HookManager.registered_handlers :my_hub, Hook.registry_pid_inserted()  \n[\n  %ProcessHub.Service.HookManager{\n    id: :handler_id_2,\n    m: MyModule,\n    f: :my_second_func,\n    a: [:hub_id, :_],\n    p: 0\n  }\n]\n```","ref":"hooks.html#listing-hook-handlers"},{"type":"extras","title":"Removing hook handlers - Hooks","doc":"You can remove a specific hook handler by calling the `cancel_handler/3` function.\n\nExample:\n```elixir\niex> alias ProcessHub.Service.HookManager\niex> alias ProcessHub.Constant.Hook \niex> HookManager.cancel_handler(:my_hub, Hook.registry_pid_inserted(), :handler_id_1)\n:ok\n```\n\nIt's also possible to dispatch hooks manually by calling the `dispatch_hook/3` or\n`dispatch_hooks/3` functions from the `ProcessHub.Service.HookManager` module.\nAlthough, it's not recommended to dispatch hooks manually as it can lead to unexpected behavior.","ref":"hooks.html#removing-hook-handlers"},{"type":"extras","title":"Available hooks - Hooks","doc":"{: .width-100}\n\n| Event Key                           | Trigger                                     | Data                                               |\n| ----------------------------------- | ------------------------------------------- | -------------------------------------------------- |\n| `:pre_cluster_join_hook`            | Node joins hub cluster, not handled         | `node()`                                           |\n| `:post_cluster_join_hook`           | Node joins hub cluster, handled             | `node()`                                           |\n| `:pre_cluster_leave_hook`           | Node leaves hub cluster, not handled        | `node()`                                           |\n| `:post_cluster_leave_hook`          | Node leaves hub cluster, handled            | `node()`                                           |\n| `:registry_pid_inserted_hook`       | Process registered                          | `{child_id(), [{node(), pid()}]}`                  |\n| `:registry_pid_removed_hook`        | Process unregistered                        | `child_id()`                                       |\n| `:children_migrated_hook`           | Processes migrated                          | `{node(), [child_spec()]}`                         |\n| `:priority_state_updated_hook`      | Priority state updated                      | `{priority_level(), list()}`                       |\n| `:pre_nodes_redistribution_hook`    | Nodes redistribution start                  | `{:nodeup or :nodedown, node()}`                   |\n| `:post_nodes_redistribution_hook`   | Nodes redistribution end                    | `{:nodeup or :nodedown, node()}`                   |\n| `:forwarded_migration_hook`         | Migration was forwarded and handled         | `[{node(), [children_start_data()]}]`              |\n| `:pre_children_start`               | Before child process is started             | `ProcessHub.Handler.ChildrenAdd.StartHandle.t()`   |\n| `:post_children_start`              | After child process is started              | `[{child_id(), pid(), node()}]`                    |\n| `:pre_children_redistribution_hook` | Before redistribution of children is called | `[{child_id(), node(), {:pid, pid()}}]`            |\n| `:coordinator_shutdown_hook`        | Coordinator shutdown                        | `any()`                                            |\n| `:process_startups_hook`            | Processes have been startups                | `[%{cid: child_id, pid: pid()}]`                   |\n| `:child_process_pid_update`         | Child process restarted with new pid        | `{node(), child_id()}`            |\n\nSee `ProcessHub.Constant.Hook` module for more information on available hooks.","ref":"hooks.html#available-hooks"},{"type":"extras","title":"Creating custom strategy","doc":"# Creating custom strategy\n\nIt's time to extend the functionality of the `ProcessHub` and we will do this\nby creating a custom strategy distribution strategy.","ref":"customstrategy.html"},{"type":"extras","title":"Implementation - Creating custom strategy","doc":"Our custom distribution strategy will be very simple and probably not very\nuseful in practice, but it will serve as a good example of how to create one.\n\nThis distribution strategy will distribute processes by comparing the node names. We will also take advance\nof using hook system to further customize the behavior of the system.\n\nLet's start by creating a new file called `compare.ex`. We can place the file anywhere\nin the project as long as it is picked up by the compiler.\n\nIt is a good practice to add some custom name spacing to the module to avoid conflicts\nwith other modules. We will use `CustomStrategy` as the namespace.\n```elixir\ndefmodule CustomStrategy.Compare do\nend\n```\n\nOur custom distribution strategy should define a struct which can contain any \nnumber of custom options. In our example, we will define a single option called\n`direction` which will determine if the selected node is either the first or the last\nnode in the list of nodes.\n\nWe will also create our implementation the `ProcessHub.Strategy.Distribution.Base` protocol.\n```elixir\ndefmodule CustomStrategy.Compare do\n\n    alias ProcessHub.Strategy.Distribution.Base, as: DistributionStrategy\n\n    # We can define custom options.\n    defstruct [\n        direction: :asc     \n    ]\n\n    defimpl DistributionStrategy, for: CustomStrategy.Compare do\n    end\nend\n```\n\nThe `ProcessHub.Strategy.Distribution.Base` protocol defines some functions that we need to implement. The first function is the `init/2` function which is called very early when\nthe coordinator is started. This can be very good place to register any custom\nhook handlers that we need to use. We will register a handler that will print the name of\nthe node that is joining the hub.\n\n```elixir\n...\ndefimpl DistributionStrategy, for: CustomStrategy.Compare do\n    alias ProcessHub.Service.HookManager\n    alias ProcessHub.Constant.Hook\n\n    @impl true\n    def init(_struct, hub_id) do\n      handler = %HookManager{\n        id: :some_id_here,\n        m: CustomStrategy.Compare,\n        f: :handle_node_join,\n        a: [hub_id, :_]\n      }\n\n      HookManager.register_handler(hub_id, Hook.pre_cluster_join(), handler)\n    end\nend\n...\n```\n\nLet's also define the handler function!\n\n```elixir\ndefmodule CustomStrategy.Compare do\n...\n    def handle_node_join(hub_id, node) do\n        IO.puts(\"Node #{node} is joining the hub #{hub_id}\")\n    end\nend\n...\n```\n\nAt this point we only have 2 more functions to implement. Let's continue by \noverriding the `children_init/4` function which does nothing but return `:ok`.\nIf we return anything else other than `:ok`, the startup will fail for this\nparticular function.\n\n```elixir\n...\n@impl true\ndef children_init(_struct, _hub_id, _child_specs, _opts), do: :ok\n...\n```\n\nWe now have the last function to implement which is the `belongs_to/4` function. This function will determine the node that the process should be distributed to. In our case, we will select a node based on it's name. \n\nThe very same function will also be used when processes are redistributed.\n\nWe also return a list of nodes although we only select one, the reason for this is that the function can return multiple nodes in case the process should be replicated.\nFor the sake of simplicity, we will ignore the replication factor in this example.\n\n```elixir\n...\n@impl true\ndef belongs_to(struct, hub_id, _child_id, _replication_factor) do\n    hub_nodes = ProcessHub.Service.Cluster.nodes(hub_id, [:include_local])\n\n    selected_node = case struct.direction do\n        :asc -> Enum.sort(hub_nodes) |> Enum.at(0)\n        :desc -> Enum.sort(hub_nodes) |> Enum.reverse() |> Enum.at(0)\n    end\n\n    [selected_node] # Ignore the replication factor and just return one node\nend\n...\n```\n\nThis is it!","ref":"customstrategy.html#implementation"},{"type":"extras","title":"Final implementation - Creating custom strategy","doc":"Our final implementation of the `CustomStrategy.Compare` module should look like this:\n\n```elixir\ndefmodule CustomStrategy.Compare do\n  alias ProcessHub.Strategy.Distribution.Base, as: DistributionStrategy\n\n  # We can define custom options.\n  defstruct [\n    direction: :asc\n  ]\n\n  def handle_node_join(hub_id, node) do\n      IO.puts(\"Node #{node} is joining the hub #{hub_id}\")\n  end\n\n  defimpl DistributionStrategy, for: CustomStrategy.Compare do\n    alias ProcessHub.Service.HookManager\n    alias ProcessHub.Constant.Hook\n\n    @impl true\n    def init(_struct, hub_id) do\n      handler = %HookManager{\n        id: :some_id_here,\n        m: CustomStrategy.Compare,\n        f: :handle_node_join,\n        a: [hub_id, :_]\n      }\n\n      HookManager.register_handler(hub_id, Hook.pre_cluster_join(), handler)\n    end\n\n    @impl true\n    def children_init(_struct, _hub_id, _child_specs, _opts), do: :ok\n\n    @impl true\n    def belongs_to(struct, hub_id, _child_id, _replication_factor) do\n        hub_nodes = ProcessHub.Service.Cluster.nodes(hub_id, [:include_local])\n\n        selected_node = case struct.direction do\n            :asc -> Enum.sort(hub_nodes) |> Enum.at(0)\n            :desc -> Enum.sort(hub_nodes) |> Enum.reverse() |> Enum.at(0)\n        end\n\n        [selected_node] # Ignore the replication factor and just return one node\n    end\n  end\nend\n```","ref":"customstrategy.html#final-implementation"},{"type":"extras","title":"Switching to the custom strategy - Creating custom strategy","doc":"Replace the default distribution strategy with our custom strategy in the `ProcessHub` configuration.\n```elixir\ndefmodule MyAp.Application do\n  use Application\n\n  @impl true\n  def start(_type, _args) do\n    children = [process_hub()]\n\n    opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n    Supervisor.start_link(children, opts)\n  end\n\n  defp process_hub() do\n    {ProcessHub, %ProcessHub{\n      hub_id: :my_hub,\n      distribution_strategy: %CustomStrategy.Compare{\n        direction: :desc\n      }\n    }}\n  end\nend\n\n```","ref":"customstrategy.html#switching-to-the-custom-strategy"},{"type":"extras","title":"Test drive - Creating custom strategy","doc":"Let's start two nodes on separate terminals.\n\n```bash\niex --name node1@127.0.0.1 --cookie mycookie -S mix\n```\n\n```bash\niex --name node2@127.0.0.1 --cookie mycookie -S mix\n```\n\nConnect the nodes by running the following command on the `node1` terminal.\n```elixir\niex> Node.connect(:\"node2@127.0.0.1\")\ntrue\n```\n\nTime to see the magic happen!\n```elixir\niex> ProcessHub.start_children(:my_hub, [\n...>    %{id: \"process1\", start: {MyProcess, :start_link, []}},\n...>    %{id: \"process2\", start: {MyProcess, :start_link, []}},\n...>    %{id: \"process3\", start: {MyProcess, :start_link, []}},\n...>    %{id: \"process4\", start: {MyProcess, :start_link, []}},\n...>    %{id: \"process5\", start: {MyProcess, :start_link, []}}\n...> ])\n{:ok, :start_initiated}\niex> ProcessHub.process_list(:my_hub, :global)\n[\n  process1: [\"node2@127.0.0.1\": #PID<23066.269.0>],\n  process2: [\"node2@127.0.0.1\": #PID<23066.270.0>],\n  process3: [\"node2@127.0.0.1\": #PID<23066.271.0>],\n  process4: [\"node2@127.0.0.1\": #PID<23066.272.0>],\n  process5: [\"node2@127.0.0.1\": #PID<23066.273.0>]\n]\n```\nWe can confirm that all processes are started on the second node if we\npass the `:desc` option to the struct.","ref":"customstrategy.html#test-drive"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}